@@ -31,38 +31,46 @@
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/464
+// https://github.com/chrisaljoudi/uBlock/issues/464
 if ( document instanceof HTMLDocument === false ) {
     //console.debug('contentscript-end.js > not a HTLMDocument');
     return false;
 }
 
+// This can happen
 if ( !vAPI ) {
     //console.debug('contentscript-end.js > vAPI not found');
     return;
 }
 
-// https://github.com/gorhill/uBlock/issues/587
+// https://github.com/chrisaljoudi/uBlock/issues/587
 // Pointless to execute without the start script having done its job.
 if ( !vAPI.contentscriptStartInjected ) {
     return;
 }
 
-// https://github.com/gorhill/uBlock/issues/456
+// https://github.com/chrisaljoudi/uBlock/issues/456
 // Already injected?
 if ( vAPI.contentscriptEndInjected ) {
     //console.debug('contentscript-end.js > content script already injected');
     return;
 }
 vAPI.contentscriptEndInjected = true;
+vAPI.styles = vAPI.styles || [];
 
 /******************************************************************************/
 
 var messager = vAPI.messaging.channel('contentscript-end.js');
 
+// https://github.com/gorhill/uMatrix/issues/144
+vAPI.shutdown.add(function() {
+    messager.close();
+});
+
+/******************************************************************************/
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/789
+// https://github.com/chrisaljoudi/uBlock/issues/789
 // Be sure that specific cosmetic filters are still applied.
 // Executed once, then flushed from memory.
 
@@ -83,16 +91,188 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
     }
 })();
 
+/******************************************************************************/
+/******************************************************************************/
+
+// https://github.com/chrisaljoudi/uBlock/issues/7
+
+var uBlockCollapser = (function() {
+    var timer = null;
+    var requestId = 1;
+    var newRequests = [];
+    var pendingRequests = {};
+    var pendingRequestCount = 0;
+    var srcProps = {
+        'embed': 'src',
+        'iframe': 'src',
+        'img': 'src',
+        'object': 'data'
+    };
+
+    var PendingRequest = function(target, tagName, attr) {
+        this.id = requestId++;
+        this.target = target;
+        this.tagName = tagName;
+        this.attr = attr;
+        pendingRequests[this.id] = this;
+        pendingRequestCount += 1;
+    };
+
+    // Because a while ago I have observed constructors are faster than
+    // literal object instanciations.
+    var BouncingRequest = function(id, tagName, url) {
+        this.id = id;
+        this.tagName = tagName;
+        this.url = url;
+        this.collapse = false;
+    };
+
+    var onProcessed = function(response) {
+        // https://github.com/gorhill/uMatrix/issues/144
+        if ( response.shutdown ) {
+            vAPI.shutdown.exec();
+            return;
+        }
+
+        var requests = response.result;
+        if ( requests === null || Array.isArray(requests) === false ) {
+            return;
+        }
+        var selectors = [];
+        var i = requests.length;
+        var request, entry, target, value;
+        while ( i-- ) {
+            request = requests[i];
+            if ( pendingRequests.hasOwnProperty(request.id) === false ) {
+                continue;
+            }
+            entry = pendingRequests[request.id];
+            delete pendingRequests[request.id];
+            pendingRequestCount -= 1;
+
+            // https://github.com/chrisaljoudi/uBlock/issues/869
+            if ( !request.collapse ) {
+                continue;
+            }
+
+            target = entry.target;
+
+            // https://github.com/chrisaljoudi/uBlock/issues/399
+            // Never remove elements from the DOM, just hide them
+            target.style.setProperty('display', 'none', 'important');
+
+            // https://github.com/chrisaljoudi/uBlock/issues/1048
+            // Use attribute to construct CSS rule
+            if ( value = target.getAttribute(entry.attr) ) {
+                selectors.push(entry.tagName + '[' + entry.attr + '="' + value + '"]');
+            }
+        }
+        if ( selectors.length !== 0 ) {
+            messager.send({
+                what: 'injectedSelectors',
+                type: 'net',
+                hostname: window.location.hostname,
+                selectors: selectors
+            });
+        }
+        // Renew map: I believe that even if all properties are deleted, an
+        // object will still use more memory than a brand new one.
+        if ( pendingRequestCount === 0 ) {
+            pendingRequests = {};
+        }
+    };
+
+    var send = function() {
+        timer = null;
+        messager.send({
+            what: 'filterRequests',
+            pageURL: window.location.href,
+            pageHostname: window.location.hostname,
+            requests: newRequests
+        }, onProcessed);
+        newRequests = [];
+    };
+
+    var process = function(delay) {
+        if ( newRequests.length === 0 ) {
+            return;
+        }
+        if ( delay === 0 ) {
+            clearTimeout(timer);
+            send();
+        } else if ( timer === null ) {
+            timer = setTimeout(send, delay || 20);
+        }
+    };
+
+    // If needed eventually, we could listen to `src` attribute changes
+    // for iframes.
+
+    var add = function(target) {
+        var tagName = target.localName;
+        var prop = srcProps[tagName];
+        if ( prop === undefined ) {
+            return;
+        }
+        // https://github.com/chrisaljoudi/uBlock/issues/174
+        // Do not remove fragment from src URL
+        var src = target[prop];
+        if ( typeof src !== 'string' || src === '' ) {
+            return;
+        }
+        if ( src.lastIndexOf('http', 0) !== 0 ) {
+            return;
+        }
+        var req = new PendingRequest(target, tagName, prop);
+        newRequests.push(new BouncingRequest(req.id, tagName, src));
+    };
+
+    var addIFrame = function(iframe) {
+        var src = iframe.src;
+        // TODO: niject content script in `about:blank` as well.
+        if ( src === '' || typeof src !== 'string' ) {
+            return;
+        }
+        if ( src.lastIndexOf('http', 0) !== 0 ) {
+            return;
+        }
+        var req = new PendingRequest(iframe, 'iframe', 'src');
+        newRequests.push(new BouncingRequest(req.id, 'iframe', src));
+    };
+
+    var iframesFromNode = function(node) {
+        if ( node.localName === 'iframe' ) {
+            addIFrame(node);
+        }
+        var iframes = node.querySelectorAll('iframe');
+        var i = iframes.length;
+        while ( i-- ) {
+            addIFrame(iframes[i]);
+        }
+        process();
+    };
+
+    return {
+        add: add,
+        addIFrame: addIFrame,
+        iframesFromNode: iframesFromNode,
+        process: process
+    };
+})();
+
+/******************************************************************************/
 /******************************************************************************/
 
 // Cosmetic filters
 
 (function() {
     if ( vAPI.skipCosmeticFiltering ) {
-        // console.debug('Abort cosmetic filtering');
+        //console.debug('Abort cosmetic filtering');
         return;
     }
 
+    //console.debug('Starts cosmetic filtering');
+
     //var timer = window.performance || Date;
     //var tStart = timer.now();
 
@@ -114,7 +294,7 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
                 },
                 retrieveHandler
             );
-            // https://github.com/gorhill/uBlock/issues/452
+            // https://github.com/chrisaljoudi/uBlock/issues/452
             retrieveHandler = nextRetrieveHandler;
         } else {
             nextRetrieveHandler(null);
@@ -122,20 +302,20 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
         lowGenericSelectors = [];
     };
 
-    // https://github.com/gorhill/uBlock/issues/452
+    // https://github.com/chrisaljoudi/uBlock/issues/452
     // This needs to be executed *after* the response from our query is
     // received, not at `DOMContentLoaded` time, or else there is a good
     // likeliness to outrun contentscript-start.js, which may still be waiting
     // on a response from its own query.
     var firstRetrieveHandler = function(response) {
-        // https://github.com/gorhill/uBlock/issues/158
+        // https://github.com/chrisaljoudi/uBlock/issues/158
         // Ensure injected styles are enforced
         // rhill 2014-11-16: not sure this is needed anymore. Test case in
         //  above issue was fine without the line below..
         var selectors = vAPI.hideCosmeticFilters;
         if ( typeof selectors === 'object' ) {
             injectedSelectors = selectors;
-            //hideElements(Object.keys(selectors));
+            hideElements(Object.keys(selectors));
         }
         // Add exception filters into injected filters collection, in order
         // to force them to be seen as "already injected".
@@ -151,24 +331,33 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
         firstRetrieveHandler = null;
 
         // These are sent only once
-        if ( response ) {
-            if ( response.highGenerics ) {
-                highGenerics = response.highGenerics;
+        var result = response && response.result;
+        if ( result ) {
+            if ( result.highGenerics ) {
+                highGenerics = result.highGenerics;
             }
-            if ( response.donthide ) {
-                processLowGenerics(response.donthide, nullArray);
+            if ( result.donthide ) {
+                processLowGenerics(result.donthide, nullArray);
             }
         }
 
         nextRetrieveHandler(response);
     };
 
-    var nextRetrieveHandler = function(selectors) {
+    var nextRetrieveHandler = function(response) {
+        // https://github.com/gorhill/uMatrix/issues/144
+        if ( response && response.shutdown ) {
+            vAPI.shutdown.exec();
+            return;
+        }
+
         //var tStart = timer.now();
         //console.debug('µBlock> contextNodes = %o', contextNodes);
+        var result = response && response.result;
         var hideSelectors = [];
-        if ( selectors && selectors.hide.length ) {
-            processLowGenerics(selectors.hide, hideSelectors);
+
+        if ( result && result.hide.length ) {
+            processLowGenerics(result.hide, hideSelectors);
         }
         if ( highGenerics ) {
             if ( highGenerics.hideLowCount ) {
@@ -196,14 +385,16 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
     // - Injecting a style tag
 
     var addStyleTag = function(selectors) {
-        //hideElements(selectors);
+        var selectorStr = selectors.toString();
         var style = document.createElement('style');
         // The linefeed before the style block is very important: do no remove!
-        style.appendChild(document.createTextNode(selectors.toString() + '\n{display:none !important;}'));
+        style.appendChild(document.createTextNode(selectorStr + '\n{display:none !important;}'));
         var parent = document.body || document.documentElement;
         if ( parent ) {
             parent.appendChild(style);
+            vAPI.styles.push(style);
         }
+        hideElements(selectorStr);
         messager.send({
             what: 'injectedSelectors',
             type: 'cosmetic',
@@ -214,7 +405,7 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
     };
 
     var hideElements = function(selectors) {
-        // https://github.com/gorhill/uBlock/issues/207
+        // https://github.com/chrisaljoudi/uBlock/issues/207
         // Do not call querySelectorAll() using invalid CSS selectors
         if ( selectors.length === 0 ) {
             return;
@@ -222,7 +413,7 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
         if ( document.body === null ) {
             return;
         }
-        // https://github.com/gorhill/uBlock/issues/158
+        // https://github.com/chrisaljoudi/uBlock/issues/158
         // Using CSSStyleDeclaration.setProperty is more reliable
         var elems = document.querySelectorAll(selectors);
         var i = elems.length;
@@ -297,7 +488,7 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
                     }
                 }
                 // Candidate 2 = specific form
-                selector = node.tagName.toLowerCase() + selector;
+                selector = node.localName + selector;
                 if ( generics.hasOwnProperty(selector) ) {
                     if ( injectedSelectors.hasOwnProperty(selector) === false ) {
                         injectedSelectors[selector] = true;
@@ -460,20 +651,22 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
         return;
     }
 
+    // https://github.com/chrisaljoudi/uBlock/issues/618
+    // Following is to observe dynamically added iframes:
+    // - On Firefox, the iframes fails to fire a `load` event
+
     var ignoreTags = {
         'link': true,
-        'LINK': true,
         'script': true,
-        'SCRIPT': true,
-        'style': true,
-        'STYLE': true
+        'style': true
     };
 
     // Added node lists will be cumulated here before being processed
     var addedNodeLists = [];
     var addedNodeListsTimer = null;
+    var collapser = uBlockCollapser;
 
-    var mutationObservedHandler = function() {
+    var treeMutationObservedHandler = function() {
         var nodeList, iNode, node;
         while ( nodeList = addedNodeLists.pop() ) {
             iNode = nodeList.length;
@@ -482,10 +675,11 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
                 if ( node.nodeType !== 1 ) {
                     continue;
                 }
-                if ( ignoreTags.hasOwnProperty(node.tagName) ) {
+                if ( ignoreTags.hasOwnProperty(node.localName) ) {
                     continue;
                 }
                 contextNodes.push(node);
+                collapser.iframesFromNode(node);
             }
         }
         addedNodeListsTimer = null;
@@ -496,7 +690,7 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
         }
     };
 
-    // https://github.com/gorhill/uBlock/issues/205
+    // https://github.com/chrisaljoudi/uBlock/issues/205
     // Do not handle added node directly from within mutation observer.
     var treeMutationObservedHandlerAsync = function(mutations) {
         var iMutation = mutations.length;
@@ -511,16 +705,26 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
             // I arbitrarily chose 100 ms for now:
             // I have to compromise between the overhead of processing too few
             // nodes too often and the delay of many nodes less often.
-            addedNodeListsTimer = setTimeout(mutationObservedHandler, 100);
+            addedNodeListsTimer = setTimeout(treeMutationObservedHandler, 100);
         }
     };
 
+    //console.debug('Starts cosmetic filtering\'s mutations observer');
+
     // https://github.com/gorhill/httpswitchboard/issues/176
     var treeObserver = new MutationObserver(treeMutationObservedHandlerAsync);
     treeObserver.observe(document.body, {
         childList: true,
         subtree: true
     });
+
+    // https://github.com/gorhill/uMatrix/issues/144
+    vAPI.shutdown.add(function() {
+        treeObserver.disconnect();
+        if ( addedNodeListsTimer !== null ) {
+            clearTimeout(addedNodeListsTimer);
+        }
+    });
 })();
 
 /******************************************************************************/
@@ -528,214 +732,57 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
 
 // Permanent
 
+// Listener to collapse blocked resources.
+// - Future requests not blocked yet
+// - Elements dynamically added to the page
+// - Elements which resource URL changes
+
 (function() {
-    // https://github.com/gorhill/uBlock/issues/683
-    // Instead of a closure we use a map to remember the element to collapse
-    var filterRequestId = 1;
-    var filterRequests = {};
-
-    var FilterRequest = function(target, selector) {
-        this.id = filterRequestId++;
-        this.target = target;
-        this.selector = selector;
-    };
-
-    FilterRequest.send = function(target, tagName, prop, src) {
-        var req = new FilterRequest(
-            target,
-            tagName + '[' + prop + '="' + src + '"]'
-        );
-        filterRequests[req.id] = req;
-        messager.send(
-            {
-                what: 'filterRequest',
-                id: req.id,
-                tagName: tagName,
-                requestURL: src,
-                pageHostname: window.location.hostname,
-                pageURL: window.location.href
-            },
-            onAnswerReceived
-        );
-    };
-
-    // Process answer: collapse, hide, or do nothing.
-
-    var onAnswerReceived = function(details) {
-        // This should not happen under normal circumstances. It probably can
-        // happen if the extension is disabled though.
-        if ( typeof details !== 'object' || details === null ) {
-            return;
-        }
-
-        // This should definitely not happen
-        if ( filterRequests.hasOwnProperty(details.id) === false ) {
-            return;
-        }
-
-        var req = filterRequests[details.id];
-        delete filterRequests[details.id];
-
-        // https://github.com/gorhill/uBlock/issues/869
-        if ( details.collapse !== true ) {
-            return;
-        }
-
-        //console.log('contentscript-end.js > onAnswerReceived(%o)', req);
-
-        // If `!important` is not there, going back using history will
-        // likely cause the hidden element to re-appear.
-        // https://github.com/gorhill/uBlock/issues/399
-
-        // Never remove elements from the DOM, just hide them
-        req.target.style.setProperty('display', 'none', 'important');
-
-        messager.send({
-            what: 'injectedSelectors',
-            type: 'net',
-            hostname: window.location.hostname,
-            selectors: req.selector
-        });
-    };
-
-    // https://github.com/gorhill/uBlock/issues/174
-    // Do not remove fragment from src URL
-
-    // TODO: Find out whether trying to send more than one filter request per
-    //       message is worth it.
-
-    var onResource = function(target, dict) {
-        if ( !target ) {
-            return;
-        }
-        var tagName = target.tagName.toLowerCase();
-        var prop = dict[tagName];
-        if ( prop === undefined ) {
-            return;
-        }
-        var src = target[prop];
-        if ( typeof src !== 'string' || src === '' ) {
-            return;
-        }
-        if ( src.lastIndexOf('http', 0) !== 0 ) {
-            return;
-        }
-        FilterRequest.send(target, tagName, prop, src);
-    };
-
-    // Listeners to mop up whatever is otherwise missed:
-    // - Future requests not blocked yet
-    // - Elements dynamically added to the page
-    // - Elements which resource URL changes
-
-    var loadedElements = {
-        'iframe': 'src'
-    };
-
-    var failedElements = {
-        'img': 'src',
-        'input': 'src',
-        'object': 'data'
-    };
-
-    var onResourceLoaded = function(ev) {
-        //console.debug('onResourceLoaded(%o)', ev);
-        onResource(ev.target, loadedElements);
-    };
-
     var onResourceFailed = function(ev) {
         //console.debug('onResourceFailed(%o)', ev);
-        onResource(ev.target, failedElements);
+        uBlockCollapser.add(ev.target);
+        uBlockCollapser.process();
     };
-
-    document.addEventListener('load', onResourceLoaded, true);
     document.addEventListener('error', onResourceFailed, true);
+
+    // https://github.com/gorhill/uMatrix/issues/144
+    vAPI.shutdown.add(function() {
+        document.removeEventListener('error', onResourceFailed, true);
+    });
 })();
 
 /******************************************************************************/
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/7
+// https://github.com/chrisaljoudi/uBlock/issues/7
 
 // Executed only once
 
 (function() {
-    var srcProps = {
-        'embed': 'src',
-        'iframe': 'src',
-        'img': 'src',
-        'object': 'data'
-    };
-    var elements = [];
+    var collapser = uBlockCollapser;
+    var elems, i, elem;
 
-    var onAnswerReceived = function(details) {
-        if ( typeof details !== 'object' || details === null ) {
-            return;
-        }
-        var collapse = details.collapse;
+    elems = document.querySelectorAll('embed, object');
+    i = elems.length;
+    while ( i-- ) {
+        collapser.add(elems[i]);
+    }
 
-        // https://github.com/gorhill/uBlock/issues/869
-        if ( collapse !== true ) {
-            return;
-        }
-
-        var requests = details.requests;
-        var selectors = [];
-        var i = requests.length;
-        var request;
-        while ( i-- ) {
-            request = requests[i];
-            // https://github.com/gorhill/uBlock/issues/399
-            // Never remove elements from the DOM, just hide them
-            elements[request.index].style.setProperty('display', 'none', 'important');
-            selectors.push(request.tagName + '[' + srcProps[request.tagName] + '="' + request.url + '"]');
-        }
-        if ( selectors.length !== 0 ) {
-            messager.send({
-                what: 'injectedSelectors',
-                type: 'net',
-                hostname: window.location.hostname,
-                selectors: selectors
-            });
-        }
-    };
-
-    var requests = [];
-    var tagNames = ['embed','iframe','img','object'];
-    var elementIndex = 0;
-    var tagName, elems, i, elem, prop, src;
-    while ( tagName = tagNames.pop() ) {
-        elems = document.getElementsByTagName(tagName);
-        i = elems.length;
-        while ( i-- ) {
-            elem = elems[i];
-            prop = srcProps[tagName];
-            if ( prop === undefined ) {
-                continue;
-            }
-            src = elem[prop];
-            if ( typeof src !== 'string' || src === '' ) {
-                continue;
-            }
-            if ( src.lastIndexOf('http', 0) !== 0 ) {
-                continue;
-            }
-            requests.push({
-                index: elementIndex,
-                tagName: tagName,
-                url: src
-            });
-            elements[elementIndex] = elem;
-            elementIndex += 1;
+    elems = document.querySelectorAll('img');
+    i = elems.length;
+    while ( i-- ) {
+        elem = elems[i];
+        if ( elem.complete ) {
+            collapser.add(elem);
         }
     }
-    var details = {
-        what: 'filterRequests',
-        pageURL: window.location.href,
-        pageHostname: window.location.hostname,
-        requests: requests
-    };
-    messager.send(details, onAnswerReceived);
+
+    elems = document.querySelectorAll('iframe');
+    i = elems.length;
+    while ( i-- ) {
+        collapser.addIFrame(elems[i]);
+    }
+    collapser.process(0);
 })();
 
 /******************************************************************************/
@@ -761,6 +808,11 @@ var messager = vAPI.messaging.channel('contentscript-end.js');
     };
 
     window.addEventListener('contextmenu', onContextMenu, true);
+
+    // https://github.com/gorhill/uMatrix/issues/144
+    vAPI.shutdown.add(function() {
+        document.removeEventListener('contextmenu', onContextMenu, true);
+    });
 })();
 
 /******************************************************************************/
