@@ -20,7 +20,7 @@
 */
 
 /* jshint bitwise: false */
-/* global µBlock */
+/* global punycode, µBlock */
 
 /******************************************************************************/
 
@@ -237,6 +237,7 @@ var FilterParser = function() {
     this.unhide = 0;
     this.hostnames = [];
     this.invalid = false;
+    this.cosmetic = true;
     this.reParser = /^\s*([^#]*)(##|#@#)(.+)\s*$/;
 };
 
@@ -248,6 +249,7 @@ FilterParser.prototype.reset = function() {
     this.unhide = 0;
     this.hostnames.length = 0;
     this.invalid = false;
+    this.cosmetic = true;
     return this;
 };
 
@@ -259,7 +261,7 @@ FilterParser.prototype.parse = function(s) {
 
     var matches = this.reParser.exec(s);
     if ( matches === null || matches.length !== 4 ) {
-        this.invalid = true;
+        this.cosmetic = false;
         return this;
     }
 
@@ -272,7 +274,7 @@ FilterParser.prototype.parse = function(s) {
     // Any sequence of `#` longer than one means the line is not a valid
     // cosmetic filter.
     if ( this.suffix.indexOf('##') !== -1 ) {
-        this.invalid = true;
+        this.cosmetic = false;
         return this;
     }
 
@@ -405,7 +407,7 @@ SelectorCacheEntry.prototype.add = function(selectors, type) {
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/420
+// https://github.com/chrisaljoudi/uBlock/issues/420
 SelectorCacheEntry.prototype.remove = function(type) {
     this.lastAccessTime = Date.now();
     if ( type === undefined || type === 'cosmetic' ) {
@@ -512,6 +514,8 @@ var FilterContainer = function() {
     this.selectorCacheAgeMax = 20 * 60 * 1000; // 20 minutes
     this.selectorCacheCountMin = 10;
     this.selectorCacheTimer = null;
+    this.reHasUnicode = /[^\x00-\x7F]/;
+    this.punycode = punycode;
     this.reset();
 };
 
@@ -557,11 +561,40 @@ FilterContainer.prototype.reset = function() {
 
 /******************************************************************************/
 
+// https://github.com/chrisaljoudi/uBlock/issues/1004
+// Detect and report invalid CSS selectors.
+
+FilterContainer.prototype.div = document.createElement('div');
+
+// Not all browsers support `Element.matches`:
+// http://caniuse.com/#feat=matchesselector
+
+if ( typeof FilterContainer.prototype.div.matches === 'function' ) {
+    FilterContainer.prototype.isValidSelector = function(s) {
+        try {
+            this.div.matches(s);
+        } catch (e) {
+            console.error('uBlock> invalid cosmetic filter:', s);
+            return false;
+        }
+        return true;
+    };
+} else {
+    FilterContainer.prototype.isValidSelector = function() {
+        return true;
+    };
+}
+
+/******************************************************************************/
+
 FilterContainer.prototype.compile = function(s, out) {
     var parsed = this.parser.parse(s);
-    if ( parsed.invalid ) {
+    if ( parsed.cosmetic === false ) {
         return false;
     }
+    if ( parsed.invalid ) {
+        return true;
+    }
 
     var hostnames = parsed.hostnames;
     var i = hostnames.length;
@@ -570,7 +603,16 @@ FilterContainer.prototype.compile = function(s, out) {
         return true;
     }
 
-    // https://github.com/gorhill/uBlock/issues/151
+    // For hostname- or entity-based filters, class- or id-based selectors are
+    // still the most common, and can easily be tested using a plain regex.
+    if (
+        this.reClassOrIdSelector.test(parsed.suffix) === false &&
+        this.isValidSelector(parsed.suffix) === false
+    ) {
+        return true;
+    }
+
+    // https://github.com/chrisaljoudi/uBlock/issues/151
     // Negated hostname means the filter applies to all non-negated hostnames
     // of same filter OR globally if there is no non-negated hostnames.
     var applyGlobally = true;
@@ -598,15 +640,13 @@ FilterContainer.prototype.compile = function(s, out) {
 FilterContainer.prototype.compileGenericSelector = function(parsed, out) {
     var selector = parsed.suffix;
 
-    // https://github.com/gorhill/uBlock/issues/497
+    // https://github.com/chrisaljoudi/uBlock/issues/497
     // All generic exception filters are put in the same bucket: they are
     // expected to be very rare.
     if ( parsed.unhide ) {
-        out.push(
-            'c\v' +
-            'g1\v' +
-            selector
-        );
+        if ( this.isValidSelector(selector) ) {
+            out.push('c\vg1\v' + selector);
+        }
         return;
     }
 
@@ -618,45 +658,56 @@ FilterContainer.prototype.compileGenericSelector = function(parsed, out) {
         if ( matches === null ) {
             return;
         }
-        out.push(
-            'c\v' +
-            (matches[1] === selector ? 'lg\v' : 'lg+\v') +
-            makeHash(0, matches[1], this.genericHashMask) + '\v' +
-            selector
-        );
+        // Single-CSS rule: no need to test for whether the selector
+        // is valid, the regex took care of this. Most generic selector falls
+        // into that category.
+        if ( matches[1] === selector ) {
+            out.push(
+                'c\vlg\v' +
+                makeHash(0, matches[1], this.genericHashMask) + '\v' +
+                selector
+            );
+            return;
+        }
+        // Many-CSS rules
+        if ( this.isValidSelector(selector) ) {
+            out.push(
+                'c\vlg+\v' +
+                makeHash(0, matches[1], this.genericHashMask) + '\v' +
+                selector
+            );
+        }
         return;
     }
 
     // ["title"] and ["alt"] will go in high-low generic bin.
     if ( this.reHighLow.test(selector) ) {
-        out.push(
-            'c\v' +
-            'hlg0\v' +
-            selector
-        );
+        if ( this.isValidSelector(selector) ) {
+            out.push('c\vhlg0\v' + selector);
+        }
         return;
     }
 
     // [href^="..."] will go in high-medium generic bin.
     matches = this.reHighMedium.exec(selector);
     if ( matches && matches.length === 2 ) {
-        out.push(
-            'c\v' +
-            'hmg0\v' +
-            matches[1] + '\v' +
-            selector
-        );
+        if ( this.isValidSelector(selector) ) {
+            out.push(
+                'c\vhmg0\v' +
+                matches[1] + '\v' +
+                selector
+            );
+        }
         return;
     }
 
     // All else
-    out.push(
-        'c\v' +
-        'hhg0\v' +
-        selector
-    );
+    if ( this.isValidSelector(selector) ) {
+        out.push('c\vhhg0\v' + selector);
+    }
 };
 
+FilterContainer.prototype.reClassOrIdSelector = /^([#.][\w-]+)$/;
 FilterContainer.prototype.rePlainSelector = /^([#.][\w-]+)/;
 FilterContainer.prototype.reHighLow = /^[a-z]*\[(?:alt|title)="[^"]+"\]$/;
 FilterContainer.prototype.reHighMedium = /^\[href\^="https?:\/\/([^"]{8})[^"]*"\]$/;
@@ -664,13 +715,20 @@ FilterContainer.prototype.reHighMedium = /^\[href\^="https?:\/\/([^"]{8})[^"]*"\
 /******************************************************************************/
 
 FilterContainer.prototype.compileHostnameSelector = function(hostname, parsed, out) {
-    // https://github.com/gorhill/uBlock/issues/145
+    // https://github.com/chrisaljoudi/uBlock/issues/145
     var unhide = parsed.unhide;
     if ( hostname.charAt(0) === '~' ) {
         hostname = hostname.slice(1);
         unhide ^= 1;
     }
-    // https://github.com/gorhill/uBlock/issues/188
+
+    // punycode if needed
+    if ( this.reHasUnicode.test(hostname) ) {
+        //console.debug('µBlock.cosmeticFilteringEngine/FilterContainer.compileHostnameSelector> punycoding:', hostname);
+        hostname = this.punycode.toASCII(hostname);
+    }
+
+    // https://github.com/chrisaljoudi/uBlock/issues/188
     // If not a real domain as per PSL, assign a synthetic one
     var hash;
     var domain = this.µburi.domainFromHostname(hostname);
@@ -796,7 +854,7 @@ FilterContainer.prototype.fromCompiledContent = function(text, lineBeg, skip) {
             continue;
         }
 
-        // https://github.com/gorhill/uBlock/issues/497
+        // https://github.com/chrisaljoudi/uBlock/issues/497
         // Generic exception filters: expected to be a rare occurrence.
         if ( fields[0] === 'g1' ) {
             this.genericDonthide.push(fields[1]);
@@ -1067,7 +1125,7 @@ FilterContainer.prototype.retrieveGenericSelectors = function(request) {
             hideHigh: this.highHighGenericHide,
             hideHighCount: this.highHighGenericHideCount
         };
-        // https://github.com/gorhill/uBlock/issues/497
+        // https://github.com/chrisaljoudi/uBlock/issues/497
         r.donthide = this.genericDonthide;
     }
 
@@ -1112,7 +1170,7 @@ FilterContainer.prototype.retrieveDomainSelectors = function(request) {
     var domain = µb.URI.domainFromHostname(hostname) || hostname;
     var pos = domain.indexOf('.');
 
-    // https://github.com/gorhill/uBlock/issues/587
+    // https://github.com/chrisaljoudi/uBlock/issues/587
     // r.ready will tell the content script the cosmetic filtering engine is
     // up and ready.
 
@@ -1132,7 +1190,7 @@ FilterContainer.prototype.retrieveDomainSelectors = function(request) {
     if ( bucket = this.hostnameFilters[hash] ) {
         bucket.retrieve(hostname, r.cosmeticHide);
     }
-    // https://github.com/gorhill/uBlock/issues/188
+    // https://github.com/chrisaljoudi/uBlock/issues/188
     // Special bucket for those filters without a valid domain name as per PSL
     if ( bucket = this.hostnameFilters[this.type0NoDomainHash] ) {
         bucket.retrieve(hostname, r.cosmeticHide);
@@ -1149,7 +1207,7 @@ FilterContainer.prototype.retrieveDomainSelectors = function(request) {
         bucket.retrieve(hostname, r.cosmeticDonthide);
     }
 
-    // https://github.com/gorhill/uBlock/issues/188
+    // https://github.com/chrisaljoudi/uBlock/issues/188
     // Special bucket for those filters without a valid domain name as per PSL
     if ( bucket = this.hostnameFilters[this.type1NoDomainHash] ) {
         bucket.retrieve(hostname, r.cosmeticDonthide);
