@@ -76,6 +76,7 @@ var nameToActionMap = {
 // For performance purpose, as simple tests as possible
 var reHostnameVeryCoarse = /[g-z_-]/;
 var reIPv4VeryCoarse = /\.\d+$/;
+var reBadHostname = /[^0-9a-z_.\[\]:-]/;
 
 // http://tools.ietf.org/html/rfc5952
 // 4.3: "MUST be represented in lowercase"
@@ -315,13 +316,18 @@ Matrix.prototype.clearRegisters = function() {
 /******************************************************************************/
 
 var is3rdParty = function(srcHostname, desHostname) {
-    if ( desHostname === '*' ) {
+    // If at least one is party-less, the relation can't be labelled
+    // "3rd-party"
+    if ( desHostname === '*' || srcHostname === '*' || srcHostname === '' ) {
         return false;
     }
-    var srcDomain = domainFromHostname(srcHostname);
-    if ( srcDomain === '' ) {
-        srcDomain = desHostname;
-    }
+
+    // No domain can very well occurs, for examples:
+    // - localhost
+    // - file-scheme
+    // etc.
+    var srcDomain = domainFromHostname(srcHostname) || srcHostname;
+
     if ( desHostname.slice(0 - srcDomain.length) !== srcDomain ) {
         return true;
     }
@@ -539,12 +545,20 @@ Matrix.prototype.fromString = function(text, append) {
         srcHostname = punycode.toASCII(fields[0]);
         desHostname = punycode.toASCII(fields[1]);
 
+        // https://github.com/chrisaljoudi/uBlock/issues/1082
+        // Discard rules with invalid hostnames
+        if ( (srcHostname !== '*' && reBadHostname.test(srcHostname)) ||
+             (desHostname !== '*' && reBadHostname.test(desHostname))
+        ) {
+            continue;
+        }
+
         type = fields[2];
         if ( typeBitOffsets.hasOwnProperty(type) === false ) {
             continue;
         }
 
-        // https://github.com/gorhill/uBlock/issues/840
+        // https://github.com/chrisaljoudi/uBlock/issues/840
         // Discard invalid rules
         if ( desHostname !== '*' && type !== '*' ) {
             continue;
