@@ -144,6 +144,10 @@ var cosmeticFilterCandidates = [];
 
 var targetElements = [];
 
+var lastNetFilterSession = window.location.host + window.location.pathname;
+var lastNetFilterHostname = '';
+var lastNetFilterUnion = '';
+
 /******************************************************************************/
 
 // For browsers not supporting `:scope`, it's not the end of the world: the
@@ -234,6 +238,76 @@ var removeElements = function(elems) {
 
 /******************************************************************************/
 
+var netFilterFromUnion = (function() {
+    var reTokenizer = /[^0-9a-z%*]+|[0-9a-z%]+|\*/gi;
+    var a = document.createElement('a');
+
+    return function(to, out) {
+        a.href= to;
+        to = a.pathname + a.search;
+        var from = lastNetFilterUnion;
+
+        // Reset reference filter when dealing with unrelated URLs
+        if ( from === '' || a.host === '' || a.host !== lastNetFilterHostname ) {
+            lastNetFilterHostname = a.host;
+            lastNetFilterUnion = to;
+            localMessager.send({
+                what: 'elementPickerEprom',
+                lastNetFilterSession: lastNetFilterSession,
+                lastNetFilterHostname: lastNetFilterHostname,
+                lastNetFilterUnion: lastNetFilterUnion
+            });
+            return;
+        }
+
+        // Related URLs
+        lastNetFilterHostname = a.host;
+
+        var fromTokens = from.match(reTokenizer);
+        var toTokens = to.match(reTokenizer);
+        var toCount = toTokens.length, toIndex = 0;
+        var fromToken, pos;
+
+        for ( var fromIndex = 0; fromIndex < fromTokens.length; fromIndex += 1 ) {
+            fromToken = fromTokens[fromIndex];
+            if ( fromToken === '*' ) {
+                continue;
+            }
+            pos = toTokens.indexOf(fromToken, toIndex);
+            if ( pos === -1 ) {
+                fromTokens[fromIndex] = '*';
+                continue;
+            }
+            if ( pos !== toIndex ) {
+                fromTokens.splice(fromIndex, 0, '*');
+                fromIndex += 1;
+            }
+            toIndex = pos + 1;
+            if ( toIndex === toCount ) {
+                fromTokens = fromTokens.slice(0, fromIndex + 1);
+                break;
+            }
+        }
+        from = fromTokens.join('').replace(/\*\*+/g, '*');
+        if ( from !== '/*' && from !== to ) {
+            out.push('||' + lastNetFilterHostname + from);
+        } else {
+            from = to;
+        }
+        lastNetFilterUnion = from;
+
+        // Remember across element picker sessions
+        localMessager.send({
+            what: 'elementPickerEprom',
+            lastNetFilterSession: lastNetFilterSession,
+            lastNetFilterHostname: lastNetFilterHostname,
+            lastNetFilterUnion: lastNetFilterUnion
+        });
+    };
+})();
+
+/******************************************************************************/
+
 // Extract the best possible net filter, i.e. as specific as possible.
 
 var netFilterFromElement = function(elem, out) {
@@ -256,15 +330,20 @@ var netFilterFromElement = function(elem, out) {
     if ( pos !== -1 ) {
         src = src.slice(0, pos);
     }
+
+    var filter = src.replace(/^https?:\/\//, '||');
+
     // Anchor absolute filter to hostname
-    src = src.replace(/^https?:\/\//, '||');
-    out.push(src);
+    out.push(filter);
+
     // Suggest a less narrow filter if possible
-    pos = src.indexOf('?');
+    pos = filter.indexOf('?');
     if ( pos !== -1 ) {
-        src = src.slice(0, pos);
-        out.push(src);
+        out.push(filter.slice(0, pos));
     }
+
+    // Suggest a filter which is a result of combining more than one URL.
+    netFilterFromUnion(src, out);
 };
 
 var netFilterSources = {
@@ -297,16 +376,17 @@ var cosmeticFilterFromElement = function(elem, out) {
     }
 
     // Class(es)
-    v = typeof elem.className === 'string' && elem.className.trim();
-    if ( v.length ) {
-        v = v.split(/\s+/);
-        i = v.length;
-        while ( i-- ) {
-            v[i] = CSS.escape(v[i]);
+    if ( suffix.length === 0 ) {
+        v = elem.classList;
+        if ( v ) {
+            i = v.length || 0;
+            while ( i-- ) {
+                suffix.push('.' + CSS.escape(v.item(i)));
+            }
         }
-        suffix.push('.', v.join('.'));
     }
 
+    // Tag name
     if ( suffix.length === 0 ) {
         prefix = tagName;
     }
@@ -348,7 +428,7 @@ var cosmeticFilterFromElement = function(elem, out) {
 
     var selector = prefix + suffix.join('');
 
-    // https://github.com/gorhill/uBlock/issues/637
+    // https://github.com/chrisaljoudi/uBlock/issues/637
     // If the selector is still ambiguous at this point, further narrow using
     // `nth-of-type`. It is preferable to use `nth-of-type` as opposed to
     // `nth-child`, as `nth-of-type` is less volatile.
@@ -381,6 +461,12 @@ var filtersFromElement = function(elem) {
         cosmeticFilterFromElement(elem, cosmeticFilterCandidates);
         elem = elem.parentNode;
     }
+    // The body tag is needed as anchor only when the immediate child
+    // uses`nth-of-type`.
+    var i = cosmeticFilterCandidates.length;
+    if ( i !== 0 && cosmeticFilterCandidates[i-1].indexOf(':nth-of-type(') !== -1 ) {
+        cosmeticFilterCandidates.push('##body');
+    }
 };
 
 /******************************************************************************/
@@ -411,7 +497,7 @@ var elementsFromFilter = function(filter) {
     // Net filters: we need to lookup manually -- translating into a
     // foolproof CSS selector is just not possible
 
-    // https://github.com/gorhill/uBlock/issues/945
+    // https://github.com/chrisaljoudi/uBlock/issues/945
     // Transform into a regular expression, this allows the user to edit and
     // insert wildcard(s) into the proposed filter
     var reStr = '';
@@ -657,7 +743,7 @@ var onSvgHovered = function(ev) {
 /******************************************************************************/
 
 var onSvgClicked = function(ev) {
-    // https://github.com/gorhill/uBlock/issues/810#issuecomment-74600694
+    // https://github.com/chrisaljoudi/uBlock/issues/810#issuecomment-74600694
     // Unpause picker if user click outside dialog
     if ( dialog.parentNode.classList.contains('paused') ) {
         unpausePicker();
@@ -691,7 +777,7 @@ var onKeyPressed = function(ev) {
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/190
+// https://github.com/chrisaljoudi/uBlock/issues/190
 // May need to dynamically adjust the height of the overlay + new position
 // of highlighted elements.
 
@@ -776,6 +862,13 @@ var startPicker = function(details) {
     pickerRoot.contentWindow.addEventListener('keydown', onKeyPressed, true);
     pickerRoot.contentWindow.focus();
 
+    // Restore net filter union data if it originate from the same URL.
+    var eprom = details.eprom || {};
+    if ( eprom.lastNetFilterSession === lastNetFilterSession ) {
+        lastNetFilterHostname = eprom.lastNetFilterHostname || '';
+        lastNetFilterUnion = eprom.lastNetFilterUnion || '';
+    }
+
     // Auto-select a specific target, if any, and if possible
 
     highlightElements([], true);
