@@ -38,7 +38,7 @@ var onMessage = function(request, sender, callback) {
     // Async
     switch ( request.what ) {
         case 'getAssetContent':
-            // https://github.com/gorhill/uBlock/issues/417
+            // https://github.com/chrisaljoudi/uBlock/issues/417
             µb.assets.get(request.url, callback);
             return;
 
@@ -76,7 +76,7 @@ var onMessage = function(request, sender, callback) {
             break;
 
         case 'reloadTab':
-            if ( vAPI.isNoTabId(request.tabId) === false ) {
+            if ( vAPI.isBehindTheSceneTabId(request.tabId) === false ) {
                 vAPI.tabs.reload(request.tabId);
                 if ( request.select && vAPI.tabs.select ) {
                     vAPI.tabs.select(request.tabId);
@@ -88,6 +88,10 @@ var onMessage = function(request, sender, callback) {
             µb.selectFilterLists(request.switches);
             break;
 
+        case 'toggleHostnameSwitch':
+            µb.toggleHostnameSwitch(request);
+            break;
+
         case 'userSettings':
             response = µb.changeUserSettings(request.name, request.value);
             break;
@@ -199,6 +203,7 @@ var getFirewallRules = function(srcHostname, desHostnames) {
 /******************************************************************************/
 
 var getStats = function(tabId, tabTitle) {
+    var tabContext = µb.tabContextManager.lookup(tabId);
     var r = {
         advancedUserEnabled: µb.userSettings.advancedUserEnabled,
         appName: vAPI.app.name,
@@ -209,36 +214,38 @@ var getStats = function(tabId, tabTitle) {
         globalAllowedRequestCount: µb.localSettings.allowedRequestCount,
         globalBlockedRequestCount: µb.localSettings.blockedRequestCount,
         netFilteringSwitch: false,
-        pageURL: '',
+        rawURL: tabContext.rawURL,
+        pageURL: tabContext.normalURL,
+        pageHostname: tabContext.rootHostname,
+        pageDomain: tabContext.rootDomain,
         pageAllowedRequestCount: 0,
         pageBlockedRequestCount: 0,
         tabId: tabId,
         tabTitle: tabTitle
     };
+
     var pageStore = µb.pageStoreFromTabId(tabId);
     if ( pageStore ) {
-        r.pageURL = pageStore.pageURL;
-        r.pageDomain = pageStore.pageDomain;
-        r.pageHostname = pageStore.pageHostname;
         r.pageBlockedRequestCount = pageStore.perLoadBlockedRequestCount;
         r.pageAllowedRequestCount = pageStore.perLoadAllowedRequestCount;
         r.netFilteringSwitch = pageStore.getNetFilteringSwitch();
         r.hostnameDict = getHostnameDict(pageStore.hostnameToCountMap);
         r.contentLastModified = pageStore.contentLastModified;
-        r.firewallRules = getFirewallRules(pageStore.pageHostname, r.hostnameDict);
-        r.canElementPicker = r.pageHostname.indexOf('.') !== -1;
+        r.firewallRules = getFirewallRules(tabContext.rootHostname, r.hostnameDict);
+        r.canElementPicker = tabContext.rootHostname.indexOf('.') !== -1;
         r.canRequestLog = canRequestLog;
+        r.noPopups = µb.hnSwitches.evaluateZ('no-popups', tabContext.rootHostname);
+        r.noStrictBlocking = µb.hnSwitches.evaluateZ('no-strict-blocking', tabContext.rootHostname);
+        r.noCosmeticFiltering = µb.hnSwitches.evaluateZ('no-cosmetic-filtering', tabContext.rootHostname);
     } else {
         r.hostnameDict = {};
         r.firewallRules = getFirewallRules();
     }
-    if ( r.pageHostname ) {
-        r.matrixIsDirty = !µb.sessionFirewall.hasSameRules(
-            µb.permanentFirewall,
-            r.pageHostname,
-            r.hostnameDict
-        );
-    }
+    r.matrixIsDirty = !µb.sessionFirewall.hasSameRules(
+        µb.permanentFirewall,
+        tabContext.rootHostname,
+        r.hostnameDict
+    );
     return r;
 };
 
@@ -277,12 +284,40 @@ var getTargetTabId = function(tab) {
 
 /******************************************************************************/
 
+var getPopupDataLazy = function(tabId, callback) {
+    var r = {
+        hiddenElementCount: ''
+    };
+    var pageStore = µb.pageStoreFromTabId(tabId);
+
+    if ( !pageStore ) {
+        callback(r);
+        return;
+    }
+
+    µb.getHiddenElementCount(tabId, function() {
+        r.hiddenElementCount = pageStore.hiddenElementCount;
+        callback(r);
+    });
+};
+
+/******************************************************************************/
+
 var onMessage = function(request, sender, callback) {
     // Async
     switch ( request.what ) {
+        case 'getPopupDataLazy':
+            getPopupDataLazy(request.tabId, callback);
+            return;
+
         case 'getPopupData':
+            if ( request.tabId === vAPI.noTabId ) {
+                callback(getStats(vAPI.noTabId, ''));
+                return;
+            }
             vAPI.tabs.get(request.tabId, function(tab) {
-                callback(getStats(getTargetTabId(tab), tab.title));
+                // https://github.com/chrisaljoudi/uBlock/issues/1012
+                callback(getStats(getTargetTabId(tab), tab ? tab.title : ''));
             });
             return;
 
@@ -423,60 +458,31 @@ var tagNameToRequestTypeMap = {
 // Evaluate many requests
 
 var filterRequests = function(pageStore, details) {
+    var requests = details.requests;
+    if ( µb.userSettings.collapseBlocked === false ) {
+        return requests;
+    }
+
+    //console.debug('messaging.js/contentscript-end.js: processing %d requests', requests.length);
+
     var µburi = µb.URI;
     var isBlockResult = µb.isBlockResult;
 
     // Create evaluation context
-    details.pageHostname = vAPI.punycodeHostname(details.pageHostname);
-    details.pageDomain = µburi.domainFromHostname(details.pageHostname);
-    details.rootHostname = pageStore.rootHostname;
-    details.rootDomain = pageStore.rootDomain;
-    details.requestHostname = '';
+    var context = pageStore.createContextFromFrameHostname(details.pageHostname);
 
-    var inRequests = details.requests;
-    var outRequests = [];
     var request;
-    var i = inRequests.length;
+    var i = requests.length;
     while ( i-- ) {
-        request = inRequests[i];
-        if ( tagNameToRequestTypeMap.hasOwnProperty(request.tagName) === false ) {
-            continue;
-        }
-        details.requestURL = vAPI.punycodeURL(request.url);
-        details.requestHostname = µburi.hostnameFromURI(details.requestURL);
-        details.requestType = tagNameToRequestTypeMap[request.tagName];
-        if ( isBlockResult(pageStore.filterRequest(details)) ) {
-            outRequests.push(request);
+        request = requests[i];
+        context.requestURL = vAPI.punycodeURL(request.url);
+        context.requestHostname = µburi.hostnameFromURI(request.url);
+        context.requestType = tagNameToRequestTypeMap[request.tagName];
+        if ( isBlockResult(pageStore.filterRequest(context)) ) {
+            request.collapse = true;
         }
     }
-    return {
-        collapse: µb.userSettings.collapseBlocked,
-        requests: outRequests
-    };
-};
-
-/******************************************************************************/
-
-// Evaluate a single request
-
-var filterRequest = function(pageStore, details) {
-    if ( tagNameToRequestTypeMap.hasOwnProperty(details.tagName) === false ) {
-        return;
-    }
-    var µburi = µb.URI;
-    details.pageHostname = vAPI.punycodeHostname(details.pageHostname);
-    details.pageDomain = µburi.domainFromHostname(details.pageHostname);
-    details.rootHostname = pageStore.rootHostname;
-    details.rootDomain = pageStore.rootDomain;
-    details.requestURL = vAPI.punycodeURL(details.requestURL);
-    details.requestHostname = µburi.hostnameFromURI(details.requestURL);
-    details.requestType = tagNameToRequestTypeMap[details.tagName];
-    if ( µb.isBlockResult(pageStore.filterRequest(details)) ) {
-        return {
-            collapse: µb.userSettings.collapseBlocked,
-            id: details.id
-        };
-    }
+    return requests;
 };
 
 /******************************************************************************/
@@ -498,8 +504,12 @@ var onMessage = function(details, sender, callback) {
 
     switch ( details.what ) {
         case 'retrieveGenericCosmeticSelectors':
-            if ( pageStore && pageStore.getGenericCosmeticFilteringSwitch() ) {
-                response = µb.cosmeticFilteringEngine.retrieveGenericSelectors(details);
+            response = {
+                shutdown: !pageStore || !pageStore.getNetFilteringSwitch(),
+                result: null
+            };
+            if ( !response.shutdown && pageStore.getGenericCosmeticFilteringSwitch() ) {
+                response.result = µb.cosmeticFilteringEngine.retrieveGenericSelectors(details);
             }
             break;
 
@@ -507,21 +517,13 @@ var onMessage = function(details, sender, callback) {
             µb.cosmeticFilteringEngine.addToSelectorCache(details);
             break;
 
-        // Evaluate many requests
         case 'filterRequests':
-            if ( pageStore && pageStore.getNetFilteringSwitch() ) {
-                response = filterRequests(pageStore, details);
-            }
-            break;
-
-        // Evaluate a single request
-        case 'filterRequest':
-            if ( pageStore && pageStore.getNetFilteringSwitch() ) {
-                // console.log('contentscript-end.js > filterRequest(%o)', details);
-                response = filterRequest(pageStore, details);
-            }
-            if ( response === undefined ) {
-                response = { id: details.id };
+            response = {
+                shutdown: !pageStore || !pageStore.getNetFilteringSwitch(),
+                result: null
+            };
+            if ( !response.shutdown ) {
+                response.result = filterRequests(pageStore, details);
             }
             break;
 
@@ -541,6 +543,56 @@ vAPI.messaging.listen('contentscript-end.js', onMessage);
 /******************************************************************************/
 /******************************************************************************/
 
+// cosmetic-*.js
+
+(function() {
+
+'use strict';
+
+/******************************************************************************/
+
+var µb = µBlock;
+
+/******************************************************************************/
+
+var onMessage = function(request, sender, callback) {
+    // Async
+    switch ( request.what ) {
+        default:
+            break;
+    }
+
+    // Sync
+    var response;
+
+    var pageStore;
+    if ( sender && sender.tab ) {
+        pageStore = µb.pageStoreFromTabId(sender.tab.id);
+    }
+
+    switch ( request.what ) {
+        case 'hiddenElementCount':
+            if ( pageStore ) {
+                pageStore.hiddenElementCount = request.count;
+            }
+            break;
+
+        default:
+            return vAPI.messaging.UNHANDLED;
+    }
+
+    callback(response);
+};
+
+vAPI.messaging.listen('cosmetic-*.js', onMessage);
+
+/******************************************************************************/
+
+})();
+
+/******************************************************************************/
+/******************************************************************************/
+
 // element-picker.js
 
 (function() {
@@ -581,7 +633,8 @@ var onMessage = function(request, sender, callback) {
                     frameContent: this.responseText.replace(reStrings, replacer),
                     target: µb.contextMenuTarget,
                     clientX: µb.contextMenuClientX,
-                    clientY: µb.contextMenuClientY
+                    clientY: µb.contextMenuClientY,
+                    eprom: µb.epickerEprom
                 });
 
                 µb.contextMenuTarget = '';
@@ -590,6 +643,7 @@ var onMessage = function(request, sender, callback) {
             };
             xhr.send();
             return;
+
         default:
             break;
     }
@@ -602,6 +656,10 @@ var onMessage = function(request, sender, callback) {
             µb.appendUserFilters(request.filters);
             break;
 
+        case 'elementPickerEprom':
+            µb.epickerEprom = request;
+            break;
+
         default:
             return vAPI.messaging.UNHANDLED;
     }
@@ -770,10 +828,44 @@ var µb = µBlock;
 
 /******************************************************************************/
 
-var getFirewallRules = function() {
+var getRules = function() {
     return {
         permanentRules: µb.permanentFirewall.toString(),
-        sessionRules: µb.sessionFirewall.toString()
+        sessionRules: µb.sessionFirewall.toString(),
+        hnSwitches: µb.hnSwitches.toString()
+    };
+};
+
+// Untangle rules and switches.
+var untangle = function(s) {
+    var textEnd = s.length;
+    var lineBeg = 0, lineEnd;
+    var line;
+    var rules = [];
+    var switches = [];
+
+    while ( lineBeg < textEnd ) {
+        lineEnd = s.indexOf('\n', lineBeg);
+        if ( lineEnd < 0 ) {
+            lineEnd = s.indexOf('\r', lineBeg);
+            if ( lineEnd < 0 ) {
+                lineEnd = textEnd;
+            }
+        }
+        line = s.slice(lineBeg, lineEnd).trim();
+        lineBeg = lineEnd + 1;
+
+        // Switches always contain a ':'
+        if ( line.indexOf(':') === -1 ) {
+            rules.push(line);
+        } else {
+            switches.push(line);
+        }
+    }
+
+    return {
+        rules: rules.join('\n'),
+        switches: switches.join('\n')
     };
 };
 
@@ -782,34 +874,40 @@ var getFirewallRules = function() {
 var onMessage = function(request, sender, callback) {
     // Async
     switch ( request.what ) {
-        default:
-            break;
+    default:
+        break;
     }
 
     // Sync
+    var r;
     var response;
 
     switch ( request.what ) {
-        case 'getFirewallRules':
-            response = getFirewallRules();
-            break;
+    case 'getFirewallRules':
+        response = getRules();
+        break;
 
-        case 'setSessionFirewallRules':
-            // https://github.com/gorhill/uBlock/issues/772
-            µb.cosmeticFilteringEngine.removeFromSelectorCache('*');
+    case 'setSessionFirewallRules':
+        // https://github.com/chrisaljoudi/uBlock/issues/772
+        µb.cosmeticFilteringEngine.removeFromSelectorCache('*');
+        r = untangle(request.rules);
+        µb.sessionFirewall.fromString(r.rules);
+        µb.hnSwitches.fromString(r.switches);
+        µb.saveHostnameSwitches();
+        response = getRules();
+        break;
 
-            µb.sessionFirewall.fromString(request.rules);
-            response = getFirewallRules();
-            break;
+    case 'setPermanentFirewallRules':
+        r = untangle(request.rules);
+        µb.permanentFirewall.fromString(r.rules);
+        µb.savePermanentFirewallRules();
+        µb.hnSwitches.fromString(r.switches);
+        µb.saveHostnameSwitches();
+        response = getRules();
+        break;
 
-        case 'setPermanentFirewallRules':
-            µb.permanentFirewall.fromString(request.rules);
-            µb.savePermanentFirewallRules();
-            response = getFirewallRules();
-            break;
-
-        default:
-            return vAPI.messaging.UNHANDLED;
+    default:
+        return vAPI.messaging.UNHANDLED;
     }
 
     callback(response);
@@ -993,8 +1091,10 @@ var backupUserData = function(callback) {
             timeStamp: Date.now(),
             version: vAPI.app.version,
             userSettings: µb.userSettings,
-            filterLists: µb.remoteBlacklists,
+            filterLists: µb.extractSelectedFilterLists(),
             netWhitelist: µb.stringFromWhitelist(µb.netWhitelist),
+            dynamicFilteringString: µb.permanentFirewall.toString(),
+            hostnameSwitchesString: µb.hnSwitches.toString(),
             userFilters: details.content
         };
         var now = new Date();
@@ -1009,7 +1109,7 @@ var backupUserData = function(callback) {
 
         µb.restoreBackupSettings.lastBackupFile = filename;
         µb.restoreBackupSettings.lastBackupTime = Date.now();
-        µb.XAL.keyvalSetMany(µb.restoreBackupSettings);
+        vAPI.storage.set(µb.restoreBackupSettings);
 
         getLocalData(callback);
     };
@@ -1021,7 +1121,7 @@ var backupUserData = function(callback) {
 
 var restoreUserData = function(request) {
     var userData = request.userData;
-    var countdown = 5;
+    var countdown = 7;
     var onCountdown = function() {
         countdown -= 1;
         if ( countdown === 0 ) {
@@ -1031,14 +1131,20 @@ var restoreUserData = function(request) {
 
     var onAllRemoved = function() {
         // Be sure to adjust `countdown` if adding/removing anything below
-        µb.XAL.keyvalSetOne('version', userData.version);
+        µb.keyvalSetOne('version', userData.version);
         µBlock.saveLocalSettings(true);
-        µb.XAL.keyvalSetMany(userData.userSettings, onCountdown);
-        µb.XAL.keyvalSetOne('remoteBlacklists', userData.filterLists, onCountdown);
-        µb.XAL.keyvalSetOne('netWhitelist', userData.netWhitelist, onCountdown);
-        µb.assets.put('assets/user/filters.txt', userData.userFilters, onCountdown);
+        vAPI.storage.set(userData.userSettings, onCountdown);
+        µb.keyvalSetOne('remoteBlacklists', userData.filterLists, onCountdown);
+        µb.keyvalSetOne('netWhitelist', userData.netWhitelist || '', onCountdown);
 
-        µb.XAL.keyvalSetMany({
+        // With versions 0.9.2.4-, dynamic rules were saved within the
+        // `userSettings` object. No longer the case.
+        var s = userData.dynamicFilteringString || userData.userSettings.dynamicFilteringString || '';
+        µb.keyvalSetOne('dynamicFilteringString', s, onCountdown);
+
+        µb.keyvalSetOne('hostnameSwitchesString', userData.hostnameSwitchesString || '', onCountdown);
+        µb.assets.put('assets/user/filters.txt', userData.userFilters, onCountdown);
+        vAPI.storage.set({
             lastRestoreFile: request.file || '',
             lastRestoreTime: Date.now(),
             lastBackupFile: '',
@@ -1046,15 +1152,19 @@ var restoreUserData = function(request) {
         }, onCountdown);
     };
 
+    // https://github.com/chrisaljoudi/uBlock/issues/1102
+    // Ensure all currently cached assets are flushed from storage AND memory.
+    µb.assets.rmrf();
+
     // If we are going to restore all, might as well wipe out clean local
     // storage
-    µb.XAL.keyvalRemoveAll(onAllRemoved);
+    vAPI.storage.clear(onAllRemoved);
 };
 
 /******************************************************************************/
 
 var resetUserData = function() {
-    µb.XAL.keyvalRemoveAll();
+    vAPI.storage.clear();
 
     // Keep global counts, people can become quite attached to numbers
     µb.saveLocalSettings(true);
@@ -1192,6 +1302,43 @@ vAPI.messaging.listen('subscriber.js', onMessage);
 
 })();
 
-// https://www.youtube.com/watch?v=3_WcygKJP1k
+/******************************************************************************/
+/******************************************************************************/
+
+// document-blocked.js
+
+(function() {
+
+'use strict';
+
+/******************************************************************************/
+
+var onMessage = function(request, sender, callback) {
+    // Async
+    switch ( request.what ) {
+        default:
+            break;
+    }
+
+    // Sync
+    var response;
+
+    switch ( request.what ) {
+        case 'temporarilyWhitelistDocument':
+            µBlock.webRequest.temporarilyWhitelistDocument(request.hostname);
+            break;
+
+        default:
+            return vAPI.messaging.UNHANDLED;
+    }
+
+    callback(response);
+};
+
+vAPI.messaging.listen('document-blocked.js', onMessage);
+
+/******************************************************************************/
+
+})();
 
 /******************************************************************************/
