@@ -384,13 +384,18 @@ NetFilteringResultCache.prototype.pruneAsyncCallback = function() {
 
 /******************************************************************************/
 
-NetFilteringResultCache.prototype.lookup = function(url) {
-    return this.urls[url];
+NetFilteringResultCache.prototype.lookup = function(context) {
+    return this.urls[context.requestType + ' ' + context.requestURL];
 };
 
 /******************************************************************************/
 /******************************************************************************/
 
+// Frame stores are used solely to associate a URL with a frame id. The
+// name `pageHostname` is used because of historical reasons. A more
+// appropriate name is `frameHostname` -- something to do in a future
+// refactoring.
+
 // To mitigate memory churning
 var frameStoreJunkyard = [];
 var frameStoreJunkyardMax = 50;
@@ -419,20 +424,13 @@ FrameStore.prototype.init = function(rootHostname, frameURL) {
     var µburi = µb.URI;
     this.pageHostname = µburi.hostnameFromURI(frameURL);
     this.pageDomain = µburi.domainFromHostname(this.pageHostname) || this.pageHostname;
-    this.rootHostname = rootHostname;
-    this.rootDomain = µburi.domainFromHostname(rootHostname) || rootHostname;
-    // This is part of the filtering evaluation context
-    this.requestURL = this.requestHostname = this.requestType = '';
-
     return this;
 };
 
 /******************************************************************************/
 
 FrameStore.prototype.dispose = function() {
-    this.pageHostname = this.pageDomain =
-    this.rootHostname = this.rootDomain =
-    this.requestURL = this.requestHostname = this.requestType = '';
+    this.pageHostname = this.pageDomain = '';
     if ( frameStoreJunkyard.length < frameStoreJunkyardMax ) {
         frameStoreJunkyard.push(this);
     }
@@ -448,39 +446,28 @@ var pageStoreJunkyardMax = 10;
 
 /******************************************************************************/
 
-var PageStore = function(tabId, pageURL) {
-    this.init(tabId, pageURL);
+var PageStore = function(tabId) {
+    this.init(tabId);
 };
 
 /******************************************************************************/
 
-PageStore.factory = function(tabId, pageURL) {
+PageStore.factory = function(tabId) {
     var entry = pageStoreJunkyard.pop();
     if ( entry === undefined ) {
-        entry = new PageStore(tabId, pageURL);
+        entry = new PageStore(tabId);
     } else {
-        entry.init(tabId, pageURL);
+        entry.init(tabId);
     }
     return entry;
 };
 
 /******************************************************************************/
 
-PageStore.prototype.init = function(tabId, pageURL) {
+PageStore.prototype.init = function(tabId) {
+    var tabContext = µb.tabContextManager.lookup(tabId);
     this.tabId = tabId;
-    this.previousPageURL = '';
-    this.pageURL = pageURL;
-    this.pageHostname = µb.URI.hostnameFromURI(pageURL);
-
-    // https://github.com/gorhill/uBlock/issues/185
-    // Use hostname if no domain can be extracted
-    this.pageDomain = µb.URI.domainFromHostname(this.pageHostname) || this.pageHostname;
-    this.rootHostname = this.pageHostname;
-    this.rootDomain = this.pageDomain;
-
-    // This is part of the filtering evaluation context
-    this.requestURL = this.requestHostname = this.requestType = '';
-
+    this.tabHostname = tabContext.rootHostname;
     this.hostnameToCountMap = {};
     this.contentLastModified = 0;
     this.frames = {};
@@ -488,13 +475,14 @@ PageStore.prototype.init = function(tabId, pageURL) {
     this.netFilteringReadTime = 0;
     this.perLoadBlockedRequestCount = 0;
     this.perLoadAllowedRequestCount = 0;
-    this.skipLocalMirroring = false;
+    this.hiddenElementCount = ''; // Empty string means "unknown"
     this.netFilteringCache = NetFilteringResultCache.factory();
 
     // Support `elemhide` filter option. Called at this point so the required
     // context is all setup at this point.
+    var context = this.createContextFromPage();
     this.skipCosmeticFiltering = µb.staticNetFilteringEngine
-                                   .matchStringExactType(this, pageURL, 'cosmetic-filtering')
+                                   .matchStringExactType(context, tabContext.normalURL, 'cosmetic-filtering')
                                    .charAt(1) === 'b';
 
     // Preserve old buffer if there is one already, it may be in use, and
@@ -508,32 +496,31 @@ PageStore.prototype.init = function(tabId, pageURL) {
 
 /******************************************************************************/
 
-PageStore.prototype.reuse = function(pageURL, context) {
+PageStore.prototype.reuse = function(context) {
+    // When force refreshing a page, the page store data needs to be reset.
+
+    // If the hostname changes, we can't merely just update the context.
+    var tabContext = µb.tabContextManager.lookup(this.tabId);
+    if ( tabContext.rootHostname !== this.tabHostname ) {
+        context = '';
+    }
+
     // If URL changes without a page reload (more and more common), then we
     // need to keep all that we collected for reuse. In particular, not
     // doing so was causing a problem in `videos.foxnews.com`: clicking a
     // video thumbnail would not work, because the frame hierarchy structure
     // was flushed from memory, while not really being flushed on the page.
     if ( context === 'tabUpdated' ) {
-        this.previousPageURL = this.pageURL;
-        this.pageURL = pageURL;
-        this.pageHostname = µb.URI.hostnameFromURI(pageURL);
-        this.pageDomain = µb.URI.domainFromHostname(this.pageHostname) || this.pageHostname;
-        this.rootHostname = this.pageHostname;
-        this.rootDomain = this.pageDomain;
-
-        // As part of https://github.com/gorhill/uBlock/issues/405
+        // As part of https://github.com/chrisaljoudi/uBlock/issues/405
         // URL changed, force a re-evaluation of filtering switch
         this.netFilteringReadTime = 0;
-
         return this;
     }
+
     // A new page is completely reloaded from scratch, reset all.
     this.disposeFrameStores();
     this.netFilteringCache = this.netFilteringCache.dispose();
-    var previousPageURL = this.pageURL;
-    this.init(this.tabId, pageURL);
-    this.previousPageURL = previousPageURL;
+    this.init(this.tabId);
     return this;
 };
 
@@ -546,10 +533,6 @@ PageStore.prototype.dispose = function() {
     // need to release the memory taken by these, which can amount to
     // sizeable enough chunks (especially requests, through the request URL
     // used as a key).
-    this.pageURL = this.previousPageURL =
-    this.pageHostname = this.pageDomain =
-    this.rootHostname = this.rootDomain =
-    this.requestURL = this.requestHostname = this.requestType = '';
     this.hostnameToCountMap = null;
     this.disposeFrameStores();
     this.netFilteringCache = this.netFilteringCache.dispose();
@@ -591,29 +574,78 @@ PageStore.prototype.setFrame = function(frameId, frameURL) {
 
 /******************************************************************************/
 
-PageStore.prototype.getNetFilteringSwitch = function() {
-    if ( this.netFilteringReadTime < µb.netWhitelistModifyTime ) {
-        this.netFiltering = µb.getNetFilteringSwitch(this.pageURL);
-        this.netFilteringReadTime = Date.now();
+PageStore.prototype.createContextFromPage = function() {
+    var context = new µb.tabContextManager.createContext(this.tabId);
+    context.pageHostname = context.rootHostname;
+    context.pageDomain = context.rootDomain;
+    return context;
+};
+
+PageStore.prototype.createContextFromFrameId = function(frameId) {
+    var context = new µb.tabContextManager.createContext(this.tabId);
+    if ( this.frames.hasOwnProperty(frameId) ) {
+        var frameStore = this.frames[frameId];
+        context.pageHostname = frameStore.pageHostname;
+        context.pageDomain = frameStore.pageDomain;
+    } else {
+        context.pageHostname = context.rootHostname;
+        context.pageDomain = context.rootDomain;
     }
+    return context;
+};
+
+PageStore.prototype.createContextFromFrameHostname = function(frameHostname) {
+    var context = new µb.tabContextManager.createContext(this.tabId);
+    context.pageHostname = frameHostname;
+    context.pageDomain = µb.URI.domainFromHostname(frameHostname) || frameHostname;
+    return context;
+};
+
+/******************************************************************************/
+
+PageStore.prototype.getNetFilteringSwitch = function() {
+    var tabContext = µb.tabContextManager.lookup(this.tabId);
+    if (
+        this.netFilteringReadTime > µb.netWhitelistModifyTime &&
+        this.netFilteringReadTime > tabContext.modifyTime
+    ) {
+        return this.netFiltering;
+    }
+
+    // https://github.com/chrisaljoudi/uBlock/issues/1078
+    // Use both the raw and normalized URLs.
+    this.netFiltering = µb.getNetFilteringSwitch(tabContext.normalURL);
+    if ( this.netFiltering && tabContext.rawURL !== tabContext.pageURL ) {
+        this.netFiltering = µb.getNetFilteringSwitch(tabContext.rawURL);
+    }
+    this.netFilteringReadTime = Date.now();
     return this.netFiltering;
 };
 
 /******************************************************************************/
 
 PageStore.prototype.getSpecificCosmeticFilteringSwitch = function() {
-    return this.getNetFilteringSwitch() &&
-           (µb.userSettings.advancedUserEnabled &&
-            µb.sessionFirewall.mustAllowCellZY(this.rootHostname, this.rootHostname, '*')) === false;
+    if ( this.getNetFilteringSwitch() === false ) {
+        return false;
+    }
+
+    var tabContext = µb.tabContextManager.lookup(this.tabId);
+
+    if ( µb.hnSwitches.evaluateZ('no-cosmetic-filtering', tabContext.rootHostname) ) {
+        return false;
+    }
+
+    return µb.userSettings.advancedUserEnabled === false ||
+           µb.sessionFirewall.mustAllowCellZY(tabContext.rootHostname, tabContext.rootHostname, '*') === false;
 };
 
 /******************************************************************************/
 
 PageStore.prototype.getGenericCosmeticFilteringSwitch = function() {
-    return this.getNetFilteringSwitch() &&
-           this.skipCosmeticFiltering === false &&
-           (µb.userSettings.advancedUserEnabled &&
-            µb.sessionFirewall.mustAllowCellZY(this.rootHostname, this.rootHostname, '*')) === false;
+    if ( this.skipCosmeticFiltering ) {
+        return false;
+    }
+    return this.getSpecificCosmeticFilteringSwitch();
 };
 
 /******************************************************************************/
@@ -634,7 +666,7 @@ PageStore.prototype.filterRequest = function(context) {
         return '';
     }
 
-    var entry = this.netFilteringCache.lookup(context.requestURL);
+    var entry = this.netFilteringCache.lookup(context);
     if ( entry !== undefined ) {
         //console.debug('cache HIT: PageStore.filterRequest("%s")', context.requestURL);
         return entry.result;
@@ -648,7 +680,11 @@ PageStore.prototype.filterRequest = function(context) {
     // We evaluate dynamic filtering first, and hopefully we can skip
     // evaluation of static filtering.
     if ( µb.userSettings.advancedUserEnabled ) {
-        var df = µb.sessionFirewall.evaluateCellZY(context.rootHostname, context.requestHostname, context.requestType);
+        var df = µb.sessionFirewall.evaluateCellZY(
+            context.rootHostname,
+            context.requestHostname,
+            context.requestType
+        );
         if ( df.mustBlockOrAllow() ) {
             result = df.toFilterString();
         }
@@ -676,7 +712,6 @@ var collapsibleRequestTypes = 'image sub_frame object';
 /******************************************************************************/
 
 PageStore.prototype.filterRequestNoCache = function(context) {
-
     if ( this.getNetFilteringSwitch() === false ) {
         return '';
     }
@@ -689,7 +724,11 @@ PageStore.prototype.filterRequestNoCache = function(context) {
     // We evaluate dynamic filtering first, and hopefully we can skip
     // evaluation of static filtering.
     if ( µb.userSettings.advancedUserEnabled ) {
-        var df = µb.sessionFirewall.evaluateCellZY(context.rootHostname, context.requestHostname, context.requestType);
+        var df = µb.sessionFirewall.evaluateCellZY(
+            context.rootHostname,
+            context.requestHostname,
+            context.requestType
+        );
         if ( df.mustBlockOrAllow() ) {
             result = df.toFilterString();
         }
@@ -711,7 +750,7 @@ PageStore.prototype.logRequest = function(context, result) {
     // be prepared to handle invalid requestHostname, I've seen this
     // happen: http://./
     if ( requestHostname === '' ) {
-        requestHostname = context.pageHostname;
+        requestHostname = context.rootHostname;
     }
     var now = Date.now();
     if ( this.hostnameToCountMap.hasOwnProperty(requestHostname) === false ) {
@@ -734,22 +773,6 @@ PageStore.prototype.logRequest = function(context, result) {
 
 /******************************************************************************/
 
-PageStore.prototype.toMirrorURL = function(requestURL) {
-    // https://github.com/gorhill/uBlock/issues/351
-    // Bypass experimental features when uBlock is disabled for a site
-    if ( µb.userSettings.experimentalEnabled === false ||
-         this.getNetFilteringSwitch() === false ||
-         this.skipLocalMirroring ) {
-        return '';
-    }
-
-    // https://code.google.com/p/chromium/issues/detail?id=387198
-    // Not all redirects will succeed, until bug above is fixed.
-    return µb.mirrors.toURL(requestURL, true);
-};
-
-/******************************************************************************/
-
 PageStore.prototype.updateBadge = function() {
     var netFiltering = this.getNetFilteringSwitch();
     var badge = '';
