@@ -29,9 +29,41 @@
 
 /******************************************************************************/
 
+// Ensure the popup is properly sized as soon as possible. It is assume the DOM
+// content is ready at this point, which should be the case given where this
+// script file is included in the HTML file.
+
+var dfPaneVisibleStored = vAPI.localStorage.getItem('popupFirewallPane') === 'true';
+
+// Hacky? I couldn't figure a CSS recipe for this problem.
+// I do not want the left pane -- optional and hidden by defaut -- to
+// dictate the height of the popup. The right pane dictates the height
+// of the popup, and the left pane will have a scrollbar if ever its
+// height is more than what is available.
+document.querySelector('#panes > div:nth-of-type(2)').style.setProperty(
+    'height',
+    document.querySelector('#panes > div:nth-of-type(1)').offsetHeight + 'px'
+);
+
+// The padlock must be manually positioned:
+// - It's vertical position depends on the height on the title bar.
+document.getElementById('saveRules').style.setProperty(
+    'top',
+    (document.getElementById('gotoPrefs').getBoundingClientRect().bottom + 4) + 'px'
+);
+
+// https://github.com/chrisaljoudi/uBlock/issues/996
+// Experimental: mitigate glitchy popup UI: immediately set the firewall pane
+// visibility to its last known state. By default the pane is hidden.
+// Will remove if it makes no difference.
+if ( dfPaneVisibleStored ) {
+    document.getElementById('panes').classList.add('dfEnabled');
+}
+
+/******************************************************************************/
+
 var popupData;
 var dfPaneBuilt = false;
-var popupHeight;
 var reIP = /^\d+(?:\.\d+){1,3}$/;
 var reSrcHostnameFromRule = /^d[abn]:([^ ]+) ([^ ]+) ([^ ]+)/;
 var scopeToSrcHostnameMap = {
@@ -51,6 +83,7 @@ var rowsToRecycle = uDom();
 var cachedPopupHash = '';
 var statsStr = vAPI.i18n('popupBlockedStats');
 var domainsHitStr = vAPI.i18n('popupHitDomainCount');
+var reNetworkRelatedURL = /^(?:ftps?|https?|wss?):\/\//;
 
 /******************************************************************************/
 
@@ -110,9 +143,11 @@ var hashFromPopupData = function(reset) {
             hasher.push(rule);
         }
     }
+    hasher.sort();
     hasher.push(uDom('body').hasClass('off'));
+    hasher.push(uDom('#no-cosmetic-filtering').hasClass('on'));
 
-    var hash = hasher.sort().join('');
+    var hash = hasher.join('');
     if ( reset ) {
         cachedPopupHash = hash;
     }
@@ -151,28 +186,16 @@ var addFirewallRow = function(des) {
     row.descendants('span:nth-of-type(1)').text(punycode.toUnicode(des));
 
     var hnDetails = popupData.hostnameDict[des] || {};
-
     var isDomain = des === hnDetails.domain;
-    row.toggleClass('isDomain', isDomain);
-    row.toggleClass('isSubDomain', !isDomain);
-    row.toggleClass('allowed', hnDetails.allowCount !== 0);
-    row.toggleClass('blocked', hnDetails.blockCount !== 0);
+    row.toggleClass('isDomain', isDomain)
+       .toggleClass('isSubDomain', !isDomain)
+       .toggleClass('allowed', hnDetails.allowCount !== 0)
+       .toggleClass('blocked', hnDetails.blockCount !== 0)
+       .toggleClass('totalAllowed', hnDetails.totalAllowCount !== 0)
+       .toggleClass('totalBlocked', hnDetails.totalBlockCount !== 0);
+
     row.appendTo('#firewallContainer');
 
-    // Hacky? I couldn't figure a CSS recipe for this problem.
-    // I do not want the left pane -- optional and hidden by defaut -- to
-    // dictate the height of the popup. The right pane dictates the height
-    // of the popup, and the left pane will have a scrollbar if ever its
-    // height is larger than what is available.
-    if ( popupHeight === undefined ) {
-        popupHeight = uDom('#panes > div:nth-of-type(1)').nodeAt(0).offsetHeight;
-        uDom('#panes > div:nth-of-type(2)').css('height', popupHeight + 'px');
-
-        // The padlock must be manually positioned:
-        // - It's vertical position depends on the height on the title bar.
-        var rect = document.getElementById('gotoPrefs').getBoundingClientRect();
-        document.getElementById('saveRules').style.setProperty('top', (rect.bottom + 4) + 'px');
-    }
     return row;
 };
 
@@ -223,7 +246,7 @@ var updateFirewallCell = function(scope, des, type, rule) {
     var aCount = hnDetails.allowCount;
     var bCount = hnDetails.blockCount;
     if ( aCount !== 0 || bCount !== 0 ) {
-        // https://github.com/gorhill/uBlock/issues/471
+        // https://github.com/chrisaljoudi/uBlock/issues/471
         aCount = Math.min(Math.ceil(Math.log(aCount + 1) / Math.LN10), 3);
         bCount = Math.min(Math.ceil(Math.log(bCount + 1) / Math.LN10), 3);
         textNode.nodeValue = threePlus.slice(0, aCount) +
@@ -241,7 +264,7 @@ var updateFirewallCell = function(scope, des, type, rule) {
     aCount = hnDetails.totalAllowCount;
     bCount = hnDetails.totalBlockCount;
     if ( aCount !== 0 || bCount !== 0 ) {
-        // https://github.com/gorhill/uBlock/issues/471
+        // https://github.com/chrisaljoudi/uBlock/issues/471
         aCount = Math.min(Math.ceil(Math.log(aCount + 1) / Math.LN10), 3);
         bCount = Math.min(Math.ceil(Math.log(bCount + 1) / Math.LN10), 3);
         textNode.nodeValue = threePlus.slice(0, aCount) +
@@ -344,14 +367,14 @@ var renderPrivacyExposure = function() {
         desHostnameDone[des] = true;
     }
 
-    // The root page domain must always be counted as connected: that's from
-    // where the root document was fetched.
-    // https://github.com/gorhill/uBlock/issues/759
-    // The root page domain must be counted if and only if it was actually
-    // obtained through a network request.
-    if ( allDomainCount !== 0 && allDomains[popupData.pageDomain] !== true ) {
-        allDomains[popupData.pageDomain] = true;
-        touchedDomainCount += 1;
+    // Domain of the page must always be included (if there is one)
+    if (
+        allDomains.hasOwnProperty(popupData.pageDomain) === false &&
+        reNetworkRelatedURL.test(popupData.rawURL)
+    ) {
+        allHostnameRows.push(popupData.pageDomain);
+        allDomains[popupData.pageDomain] = false;
+        allDomainCount += 1;
     }
 
     var summary = domainsHitStr.replace('{{count}}', touchedDomainCount.toLocaleString())
@@ -408,10 +431,24 @@ var renderPopup = function() {
     // This will collate all domains, touched or not
     renderPrivacyExposure();
 
-    // https://github.com/gorhill/uBlock/issues/470
+    // Extra tools
+    uDom('#no-popups').toggleClass('on', popupData.noPopups === true);
+    uDom('#no-strict-blocking').toggleClass('on', popupData.noStrictBlocking === true);
+    uDom('#no-cosmetic-filtering').toggleClass('on', popupData.noCosmeticFiltering === true);
+
+    // https://github.com/chrisaljoudi/uBlock/issues/470
     // This must be done here, to be sure the popup is resized properly
     var dfPaneVisible = popupData.dfEnabled && popupData.advancedUserEnabled;
 
+    // https://github.com/chrisaljoudi/uBlock/issues/1068
+    // Remember the last state of the firewall pane. This allows to
+    // configure the popup size early next time it is opened, which means a
+    // less glitchy popup at open time.
+    if ( dfPaneVisible !== dfPaneVisibleStored ) {
+        dfPaneVisibleStored = dfPaneVisible;
+        vAPI.localStorage.setItem('popupFirewallPane', dfPaneVisibleStored);
+    }
+
     uDom('#panes').toggleClass('dfEnabled', dfPaneVisible);
     uDom('#firewallContainer').toggleClass('minimized', popupData.firewallPaneMinimized);
 
@@ -423,6 +460,22 @@ var renderPopup = function() {
 
 /******************************************************************************/
 
+var renderPopupLazy = function() {
+    var onDataReady = function(data) {
+        var v = data.hiddenElementCount || '';
+        uDom('#no-cosmetic-filtering > span.badge').text(
+            typeof v === 'number' ? v.toLocaleString() : v
+        );
+    };
+
+    messager.send({
+        what: 'getPopupDataLazy',
+        tabId: popupData.tabId
+    }, onDataReady);
+};
+
+/******************************************************************************/
+
 var toggleNetFilteringSwitch = function(ev) {
     if ( !popupData || !popupData.pageURL ) {
         return;
@@ -488,6 +541,13 @@ var toggleFirewallPane = function() {
         value: popupData.dfEnabled
     });
 
+    // https://github.com/chrisaljoudi/uBlock/issues/996
+    // Remember the last state of the firewall pane. This allows to
+    // configure the popup size early next time it is opened, which means a
+    // less glitchy popup at open time.
+    dfPaneVisibleStored = popupData.dfEnabled;
+    vAPI.localStorage.setItem('popupFirewallPane', dfPaneVisibleStored);
+
     // Dynamic filtering pane may not have been built yet
     uDom('#panes').toggleClass('dfEnabled', popupData.dfEnabled);
     if ( popupData.dfEnabled && dfPaneBuilt === false ) {
@@ -579,7 +639,7 @@ var reloadTab = function() {
 
     // Polling will take care of refreshing the popup content
 
-    // https://github.com/gorhill/uBlock/issues/748
+    // https://github.com/chrisaljoudi/uBlock/issues/748
     // User forces a reload, assume the popup has to be updated regardless if
     // there were changes or not.
     popupData.contentLastModified = -1;
@@ -614,6 +674,25 @@ var saveFirewallRules = function() {
 
 /******************************************************************************/
 
+var toggleHostnameSwitch = function() {
+    var elem = uDom(this);
+    var switchName = elem.attr('id');
+    if ( !switchName ) {
+        return;
+    }
+    elem.toggleClass('on');
+    messager.send({
+        what: 'toggleHostnameSwitch',
+        name: switchName,
+        hostname: popupData.pageHostname,
+        state: elem.hasClass('on'),
+        tabId: popupData.tabId
+    });
+    hashFromPopupData();
+};
+
+/******************************************************************************/
+
 // Poll for changes.
 //
 // I couldn't find a better way to be notified of changes which can affect
@@ -671,6 +750,7 @@ var getPopupData = function(tabId) {
     var onDataReceived = function(response) {
         cachePopupData(response);
         renderPopup();
+        renderPopupLazy(); // low priority rendering
         hashFromPopupData(true);
         pollForContentChange();
     };
@@ -679,6 +759,39 @@ var getPopupData = function(tabId) {
 
 /******************************************************************************/
 
+var onShowTooltip = function() {
+    if ( popupData.advancedUserEnabled ) {
+        return;
+    }
+
+    var tip = document.getElementById('tooltip');
+    var target = this;
+
+    tip.textContent = target.getAttribute('data-tip');
+    tip.style.removeProperty('top');
+    tip.style.removeProperty('bottom');
+
+    // Default is "over"
+    var pos;
+    var over = target.getAttribute('data-tip-position') !== 'under';
+    if ( over ) {
+        pos = document.body.getBoundingClientRect().height -
+              target.getBoundingClientRect().top;
+        tip.style.setProperty('bottom', pos + 'px');
+    } else {
+        pos = target.getBoundingClientRect().bottom;
+        tip.style.setProperty('top', pos + 'px');
+    }
+
+    uDom(tip).addClass('show');
+};
+
+var onHideTooltip = function() {
+    uDom('#tooltip').removeClass('show');
+};
+
+/******************************************************************************/
+
 // Make menu only when popup html is fully loaded
 
 uDom.onLoad(function () {
@@ -696,8 +809,12 @@ uDom.onLoad(function () {
     uDom('a[href]').on('click', gotoURL);
     uDom('h2').on('click', toggleFirewallPane);
     uDom('#refresh').on('click', reloadTab);
+    uDom('.hnSwitch').on('click', toggleHostnameSwitch);
     uDom('#saveRules').on('click', saveFirewallRules);
     uDom('[data-i18n="popupAnyRulePrompt"]').on('click', toggleMinimize);
+
+    uDom('body').on('mouseenter', '[data-tip]', onShowTooltip)
+                .on('mouseleave', '[data-tip]', onHideTooltip);
 });
 
 /******************************************************************************/
