@@ -54,14 +54,15 @@ var AnyParty = 0 << 2;
 var FirstParty = 1 << 2;
 var ThirdParty = 2 << 2;
 
-var AnyType = 1 << 4;
+var AnyType = 0 << 4;
 var typeNameToTypeValue = {
-        'stylesheet':  2 << 4,
-             'image':  3 << 4,
-            'object':  4 << 4,
-            'script':  5 << 4,
-    'xmlhttprequest':  6 << 4,
-         'sub_frame':  7 << 4,
+        'stylesheet':  1 << 4,
+             'image':  2 << 4,
+            'object':  3 << 4,
+            'script':  4 << 4,
+    'xmlhttprequest':  5 << 4,
+         'sub_frame':  6 << 4,
+              'font':  7 << 4,
              'other':  8 << 4,
 'cosmetic-filtering': 13 << 4,
      'inline-script': 14 << 4,
@@ -71,13 +72,9 @@ var typeOtherValue = typeNameToTypeValue.other;
 
 // All network request types to bitmap
 //   bring origin to 0 (from 4 -- see typeNameToTypeValue)
-//   add 2 = number of left shift to use
 //   left-shift 1 by the above-calculated value
-//   subtract 4 to set all type bits, *except* for 2 lsb
-
-// https://github.com/gorhill/uBlock/issues/723
-// The 2 lsb *must* be zeroed
-var allNetRequestTypesBitmap = (1 << (typeOtherValue >>> 4) + 2) - 4;
+//   subtract 1 to set all type bits
+var allNetRequestTypesBitmap = (1 << (typeOtherValue >>> 4)) - 1;
 
 var BlockAnyTypeAnyParty = BlockAction | AnyType | AnyParty;
 var BlockAnyType = BlockAction | AnyType;
@@ -100,6 +97,8 @@ var reURLPostHostnameAnchors = /[\/?#]/;
 
 var pageHostnameRegister = '';
 var requestHostnameRegister = '';
+//var filterRegister = null;
+//var categoryRegister = '';
 
 /******************************************************************************/
 
@@ -165,16 +164,34 @@ var isFirstParty = function(firstPartyDomain, hostname) {
     return c === '.' || c === '';
 };
 
-// https://developer.mozilla.org/en/docs/Web/JavaScript/Guide/Regular_Expressions
+var alwaysTruePseudoRegex = {
+    match: { '0': '', index: 0 },
+    exec: function(s) {
+        this.match['0'] = s;
+        return this.match;
+    },
+    test: function() {
+        return true;
+    }
+};
 
 var strToRegex = function(s, anchor) {
+    // https://github.com/chrisaljoudi/uBlock/issues/1038
+    // Special case: always match.
+    if ( s === '*' ) {
+        return alwaysTruePseudoRegex;
+    }
+
+    // https://developer.mozilla.org/en/docs/Web/JavaScript/Guide/Regular_Expressions
     var reStr = s.replace(/[.+?^${}()|[\]\\]/g, '\\$&')
                  .replace(/\*/g, '.*');
+
     if ( anchor < 0 ) {
         reStr = '^' + reStr;
     } else if ( anchor > 0 ) {
         reStr += reStr + '$';
     }
+
     //console.debug('µBlock.staticNetFilteringEngine: created RegExp("%s")', reStr);
     return new RegExp(reStr);
 };
@@ -516,7 +533,7 @@ FilterPlainRightAnchoredHostname.fromSelfie = function(s) {
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/235
+// https://github.com/chrisaljoudi/uBlock/issues/235
 // The filter is left-anchored somewhere within the hostname part of the URL.
 
 var FilterPlainHnAnchored = function(s) {
@@ -1082,7 +1099,7 @@ var getFilterClass = function(details) {
         return FilterRegex;
     }
     var s = details.f;
-    if ( s.indexOf('*') !== -1 ) {
+    if ( s.indexOf('*') !== -1 || details.token === '*' ) {
         if ( details.hostnameAnchored ) {
             return FilterGenericHnAnchored;
         }
@@ -1113,7 +1130,7 @@ var getHostnameBasedFilterClass = function(details) {
         return FilterRegexHostname;
     }
     var s = details.f;
-    if ( s.indexOf('*') !== -1 ) {
+    if ( s.indexOf('*') !== -1 || details.token === '*' ) {
         if ( details.hostnameAnchored ) {
             return FilterGenericHnAnchoredHostname;
         }
@@ -1178,6 +1195,7 @@ FilterParser.prototype.toNormalizedType = {
             'script': 'script',
     'xmlhttprequest': 'xmlhttprequest',
        'subdocument': 'sub_frame',
+              'font': 'font',
              'other': 'other',
           'elemhide': 'cosmetic-filtering',
      'inline-script': 'inline-script',
@@ -1210,14 +1228,14 @@ FilterParser.prototype.reset = function() {
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/589
+// https://github.com/chrisaljoudi/uBlock/issues/589
 // Be ready to handle multiple negated types
 
 FilterParser.prototype.parseOptType = function(raw, not) {
-    var type = typeNameToTypeValue[this.toNormalizedType[raw]];
+    var typeBit = 1 << ((typeNameToTypeValue[this.toNormalizedType[raw]] >>> 4) - 1);
 
     if ( !not ) {
-        this.types |= 1 << (type >>> 4);
+        this.types |= typeBit;
         return;
     }
 
@@ -1226,7 +1244,7 @@ FilterParser.prototype.parseOptType = function(raw, not) {
         this.types = allNetRequestTypesBitmap;
     }
 
-    this.types &= ~(1 << (type >>> 4));
+    this.types &= ~typeBit;
 };
 
 /******************************************************************************/
@@ -1270,10 +1288,14 @@ FilterParser.prototype.parseOptions = function(s) {
             this.parseOptParty(not);
             continue;
         }
-        if ( opt === 'elemhide' && this.action === AllowAction ) {
-            this.parseOptType('elemhide', false);
-            this.action = BlockAction;
-            continue;
+        if ( opt === 'elemhide' ) {
+            if ( this.action !== AllowAction ) {
+                this.parseOptType('elemhide', false);
+                this.action = BlockAction;
+                continue;
+            }
+            this.unsupported = true;
+            break;
         }
         if ( this.toNormalizedType.hasOwnProperty(opt) ) {
             this.parseOptType(opt, not);
@@ -1356,6 +1378,12 @@ FilterParser.prototype.parse = function(raw) {
                 //console.debug('µBlock.staticNetFilteringEngine/FilterParser.parse():', raw, '=', s);
             }
         }
+
+        // https://github.com/chrisaljoudi/uBlock/issues/1096
+        if ( s.charAt(0) === '^' ) {
+            this.unsupported = true;
+            return this;
+        }
     }
 
     // left-anchored
@@ -1380,8 +1408,7 @@ FilterParser.prototype.parse = function(raw) {
 
     // nothing left?
     if ( s === '' ) {
-        this.unsupported = true;
-        return this;
+        s = '*';
     }
 
     // plain hostname?
@@ -1455,27 +1482,37 @@ FilterParser.prototype.makeToken = function() {
         return;
     }
 
+    var s = this.f;
+
+    // https://github.com/chrisaljoudi/uBlock/issues/1038
+    // Match any URL.
+    if ( s === '*' ) {
+        this.token = '*';
+        return;
+    }
+
     var matches;
 
     // Hostname-anchored with no wildcard always have a token index of 0.
-    if ( this.hostnameAnchored && this.f.indexOf('*') === -1 ) {
-        matches = findHostnameToken(this.f);
+    if ( this.hostnameAnchored && s.indexOf('*') === -1 ) {
+        matches = findHostnameToken(s);
         if ( !matches || matches[0].length === 0 ) {
             return;
         }
         this.tokenBeg = matches.index;
         this.tokenEnd = reHostnameToken.lastIndex;
-        this.token = this.f.slice(this.tokenBeg, this.tokenEnd);
+        this.token = s.slice(this.tokenBeg, this.tokenEnd);
         return;
     }
 
-    matches = findFirstGoodToken(this.f);
+    matches = findFirstGoodToken(s);
     if ( matches === null || matches[0].length === 0 ) {
+        this.token = '*';
         return;
     }
     this.tokenBeg = matches.index;
     this.tokenEnd = reGoodToken.lastIndex;
-    this.token = this.f.slice(this.tokenBeg, this.tokenEnd);
+    this.token = s.slice(this.tokenBeg, this.tokenEnd);
 };
 
 /******************************************************************************/
@@ -1692,7 +1729,7 @@ FilterContainer.prototype.compile = function(raw, out) {
     }
 
     // Pure hostnames, use more efficient liquid dict
-    // https://github.com/gorhill/uBlock/issues/665
+    // https://github.com/chrisaljoudi/uBlock/issues/665
     // Create a dict keyed on request type etc.
     if ( parsed.hostnamePure && this.compileHostnameOnlyFilter(parsed, out) ) {
         return true;
@@ -1721,9 +1758,20 @@ FilterContainer.prototype.compileHostnameOnlyFilter = function(parsed, out) {
         party = parsed.firstParty ? FirstParty : ThirdParty;
     }
     var keyShard = parsed.action | parsed.important | party;
-    var type = parsed.types >>> 1 || 1; // bit 0 is unused; also, default to AnyType
+
+    var type = parsed.types;
+    if ( type === 0 ) {
+        out.push(
+            'n\v' +
+            this.makeCategoryKey(keyShard) + '\v' +
+            '.\v' +
+            parsed.f
+        );
+        return true;
+    }
+
     var bitOffset = 1;
-    while ( type !== 0 ) {
+    do {
         if ( type & 1 ) {
             out.push(
                 'n\v' +
@@ -1734,7 +1782,7 @@ FilterContainer.prototype.compileHostnameOnlyFilter = function(parsed, out) {
         }
         bitOffset += 1;
         type >>>= 1;
-    }
+    } while ( type !== 0 );
     return true;
 };
 
@@ -1790,7 +1838,7 @@ FilterContainer.prototype.compileFilter = function(parsed, out) {
         if ( filterClass === null ) {
             return false;
         }
-        // https://github.com/gorhill/uBlock/issues/251
+        // https://github.com/chrisaljoudi/uBlock/issues/251
         // Apply third-party option if it is present
         this.compileToAtomicFilter(filterClass, parsed, party, out);
     }
@@ -1808,7 +1856,7 @@ FilterContainer.prototype.compileFilter = function(parsed, out) {
         if ( filterClass === null ) {
             return false;
         }
-        // https://github.com/gorhill/uBlock/issues/191#issuecomment-53654024
+        // https://github.com/chrisaljoudi/uBlock/issues/191#issuecomment-53654024
         // If it is a block filter, we need to reverse the order of
         // evaluation.
         if ( parsed.action === BlockAction ) {
@@ -1823,9 +1871,19 @@ FilterContainer.prototype.compileFilter = function(parsed, out) {
 
 FilterContainer.prototype.compileToAtomicFilter = function(filterClass, parsed, party, out, hostname) {
     var bits = parsed.action | parsed.important | party;
-    var type = parsed.types >>> 1 || 1; // bit 0 is unused; also, default to AnyType
+    var type = parsed.types;
+    if ( type === 0 ) {
+        out.push(
+            'n\v' +
+            this.makeCategoryKey(bits) + '\v' +
+            parsed.token + '\v' +
+            filterClass.fid + '\v' +
+            filterClass.compile(parsed, hostname)
+        );
+        return;
+    }
     var bitOffset = 1;
-    while ( type !== 0 ) {
+    do {
         if ( type & 1 ) {
             out.push(
                 'n\v' +
@@ -1837,7 +1895,7 @@ FilterContainer.prototype.compileToAtomicFilter = function(filterClass, parsed,
         }
         bitOffset += 1;
         type >>>= 1;
-    }
+    } while ( type !== 0 );
 };
 
 /******************************************************************************/
@@ -1926,7 +1984,14 @@ FilterContainer.prototype.tokenize = function(url) {
         tokenEntry.beg = matches.index;
         tokenEntry.token = matches[0];
         i += 1;
+
+        // https://github.com/chrisaljoudi/uBlock/issues/1118
+        // Crazy case... but I guess we have to expect the worst...
+        if ( i === 2048 ) {
+            break;
+        }
     }
+
     // Sentinel
     tokenEntry = tokens[i];
     if ( tokenEntry === undefined ) {
@@ -1972,7 +2037,7 @@ FilterContainer.prototype.matchTokens = function(bucket, url) {
 
 // Specialized handlers
 
-// https://github.com/gorhill/uBlock/issues/116
+// https://github.com/chrisaljoudi/uBlock/issues/116
 // Some type of requests are exceptional, they need custom handling,
 // not the generic handling.
 
@@ -1986,14 +2051,18 @@ FilterContainer.prototype.matchStringExactType = function(context, requestURL, r
     var party = isFirstParty(context.pageDomain, requestHostnameRegister) ? FirstParty : ThirdParty;
 
     // Be prepared to support unknown types
-    var type = typeNameToTypeValue[requestType] || typeOtherValue;
+    var type = typeNameToTypeValue[requestType] || 0;
+    if ( type === 0 ) {
+        return '';
+    }
+
     var categories = this.categories;
     var bf = false, bucket;
 
     // Tokenize only once
     this.tokenize(url);
 
-    // https://github.com/gorhill/uBlock/issues/139
+    // https://github.com/chrisaljoudi/uBlock/issues/139
     // Test against important block filters
     if ( bucket = categories[this.makeCategoryKey(BlockAnyParty | Important | type)] ) {
         bf = this.matchTokens(bucket, url);
@@ -2043,11 +2112,11 @@ FilterContainer.prototype.matchStringExactType = function(context, requestURL, r
 /******************************************************************************/
 
 FilterContainer.prototype.matchString = function(context) {
-    // https://github.com/gorhill/uBlock/issues/519
+    // https://github.com/chrisaljoudi/uBlock/issues/519
     // Use exact type match for anything beyond `other`
     // Also, be prepared to support unknown types
     var type = typeNameToTypeValue[context.requestType] || typeOtherValue;
-    if ( type > 8 << 4 ) {
+    if ( type > typeOtherValue ) {
         return this.matchStringExactType(context, context.requestURL, context.requestType);
     }
 
@@ -2092,7 +2161,7 @@ FilterContainer.prototype.matchString = function(context) {
 
     var bf = false;
 
-    // https://github.com/gorhill/uBlock/issues/139
+    // https://github.com/chrisaljoudi/uBlock/issues/139
     // Test against important block filters.
     // The purpose of the `important` option is to reverse the order of
     // evaluation. Normally, it is "evaluate block then evaluate allow", with
