@@ -38,6 +38,14 @@
 
 /******************************************************************************/
 
+µBlock.keyvalSetOne = function(key, val, callback) {
+    var bin = {};
+    bin[key] = val;
+    vAPI.storage.set(bin, callback || this.noopFunc);
+};
+
+/******************************************************************************/
+
 µBlock.saveLocalSettings = function(force) {
     if ( force ) {
         this.localSettingsModifyTime = Date.now();
@@ -70,8 +78,13 @@
 /******************************************************************************/
 
 µBlock.savePermanentFirewallRules = function() {
-    this.userSettings.dynamicFilteringString = this.permanentFirewall.toString();
-    this.XAL.keyvalSetOne('dynamicFilteringString', this.userSettings.dynamicFilteringString);
+    this.keyvalSetOne('dynamicFilteringString', this.permanentFirewall.toString());
+};
+
+/******************************************************************************/
+
+µBlock.saveHostnameSwitches = function() {
+    this.keyvalSetOne('hostnameSwitchesString', this.hnSwitches.toString());
 };
 
 /******************************************************************************/
@@ -86,6 +99,26 @@
 
 /******************************************************************************/
 
+// This will remove all unused filter list entries from
+// µBlock.remoteBlacklists`. This helps reduce the size of backup files.
+
+µBlock.extractSelectedFilterLists = function() {
+    var r = JSON.parse(JSON.stringify(this.remoteBlacklists));
+
+    for ( var path in r ) {
+        if ( r.hasOwnProperty(path) === false ) {
+            continue;
+        }
+        if ( r[path].off !== false ) {
+            delete r[path];
+        }
+    }
+
+    return r;
+};
+
+/******************************************************************************/
+
 µBlock.saveUserFilters = function(content, callback) {
     return this.assets.put(this.userFiltersPath, content, callback);
 };
@@ -129,7 +162,7 @@
         if ( details.error ) {
             return;
         }
-        // https://github.com/gorhill/uBlock/issues/976
+        // https://github.com/chrisaljoudi/uBlock/issues/976
         // If we reached this point, the filter quite probably needs to be
         // added for sure: do not try to be too smart, trying to avoid
         // duplicates at this point may lead to more issues.
@@ -146,7 +179,7 @@
     var relocationMap = {};
 
     var fixLocation = function(location) {
-        // https://github.com/gorhill/uBlock/issues/418
+        // https://github.com/chrisaljoudi/uBlock/issues/418
         // We now support built-in external filter lists
         if ( /^https?:/.test(location) === false ) {
             location = 'assets/thirdparties/' + location;
@@ -183,7 +216,7 @@
             }
             // This may happen if the list name was pulled from the list
             // content.
-            // https://github.com/gorhill/uBlock/issues/982
+            // https://github.com/chrisaljoudi/uBlock/issues/982
             // There is no guarantee the title was successfully extracted from
             // the list content.
             if ( availableEntry.title === '' &&
