@@ -55,6 +55,8 @@ vAPI.app.restart = function() {
 
 /******************************************************************************/
 
+// chrome.storage.local.get(null, function(bin){ console.debug('%o', bin); });
+
 vAPI.storage = chrome.storage.local;
 
 /******************************************************************************/
@@ -63,7 +65,7 @@ vAPI.tabs = {};
 
 /******************************************************************************/
 
-vAPI.isNoTabId = function(tabId) {
+vAPI.isBehindTheSceneTabId = function(tabId) {
     return tabId.toString() === '-1';
 };
 
@@ -71,33 +73,124 @@ vAPI.noTabId = '-1';
 
 /******************************************************************************/
 
-var onCreatedNavigationTarget = function(details) {
-    vAPI.tabs.onPopup({
-        openerTabId: details.sourceTabId,
-        openerURL: '',
-        targetURL: details.url,
-        targetTabId: details.tabId
-    });
-};
-
-/******************************************************************************/
-
 vAPI.tabs.registerListeners = function() {
-    if ( typeof this.onNavigation === 'function' ) {
-        chrome.webNavigation.onCommitted.addListener(this.onNavigation);
-    }
+    var onNavigationClient = this.onNavigation || noopFunc;
+    var onPopupClient = this.onPopup || noopFunc;
+    var onUpdatedClient = this.onUpdated || noopFunc;
 
-    if ( typeof this.onUpdated === 'function' ) {
-        chrome.tabs.onUpdated.addListener(this.onUpdated);
-    }
+    // https://developer.chrome.com/extensions/webNavigation
+    // [onCreatedNavigationTarget ->]
+    //  onBeforeNavigate ->
+    //  onCommitted ->
+    //  onDOMContentLoaded ->
+    //  onCompleted
+
+    var popupCandidates = Object.create(null);
+
+    var PopupCandidate = function(details) {
+        this.targetTabId = details.tabId.toString();
+        this.openerTabId = details.sourceTabId.toString();
+        this.targetURL = details.url;
+        this.selfDestructionTimer = null;
+    };
+
+    PopupCandidate.prototype.selfDestruct = function() {
+        if ( this.selfDestructionTimer !== null ) {
+            clearTimeout(this.selfDestructionTimer);
+        }
+        delete popupCandidates[this.targetTabId];
+    };
+
+    PopupCandidate.prototype.launchSelfDestruction = function() {
+        if ( this.selfDestructionTimer !== null ) {
+            clearTimeout(this.selfDestructionTimer);
+        }
+        this.selfDestructionTimer = setTimeout(this.selfDestruct.bind(this), 10000);
+    };
+
+    var popupCandidateCreate = function(details) {
+        var popup = popupCandidates[details.tabId];
+        // This really should not happen...
+        if ( popup !== undefined ) {
+            return;
+        }
+        return popupCandidates[details.tabId] = new PopupCandidate(details);
+    };
+
+    var popupCandidateTest = function(details) {
+        var popup = popupCandidates[details.tabId];
+        if ( popup === undefined ) {
+            return;
+        }
+        popup.targetURL = details.url;
+        if ( onPopupClient(popup) !== true ) {
+            return;
+        }
+        popup.selfDestruct();
+        return true;
+    };
+
+    var popupCandidateDestroy = function(details) {
+        var popup = popupCandidates[details.tabId];
+        if ( popup instanceof PopupCandidate ) {
+            popup.launchSelfDestruction();
+        }
+    };
+
+    // The chrome.webRequest.onBeforeRequest() won't be called for everything
+    // else than `http`/`https`. Thus, in such case, we will bind the tab as
+    // early as possible in order to increase the likelihood of a context
+    // properly setup if network requests are fired from within the tab.
+    // Example: Chromium + case #6 at
+    //          http://raymondhill.net/ublock/popup.html
+    var reGoodForWebRequestAPI = /^https?:\/\//;
+
+    var onCreatedNavigationTarget = function(details) {
+        //console.debug('onCreatedNavigationTarget: popup candidate tab id %d = "%s"', details.tabId, details.url);
+        if ( reGoodForWebRequestAPI.test(details.url) === false ) {
+            details.frameId = 0;
+            onNavigationClient(details);
+        }
+        popupCandidateCreate(details);
+        popupCandidateTest(details);
+    };
+
+    var onBeforeNavigate = function(details) {
+        if ( details.frameId !== 0 ) {
+            return;
+        }
+        //console.debug('onBeforeNavigate: popup candidate tab id %d = "%s"', details.tabId, details.url);
+        popupCandidateTest(details);
+    };
+
+    var onUpdated = function(tabId, changeInfo, tab) {
+        if ( changeInfo.url && popupCandidateTest({ tabId: tabId, url: changeInfo.url }) ) {
+            return;
+        }
+        onUpdatedClient(tabId, changeInfo, tab);
+    };
+
+    var onCommitted = function(details) {
+        if ( details.frameId !== 0 ) {
+            return;
+        }
+        onNavigationClient(details);
+        //console.debug('onCommitted: popup candidate tab id %d = "%s"', details.tabId, details.url);
+        if ( popupCandidateTest(details) === true ) {
+            return;
+        }
+        popupCandidateDestroy(details);
+    };
+
+    chrome.webNavigation.onCreatedNavigationTarget.addListener(onCreatedNavigationTarget);
+    chrome.webNavigation.onBeforeNavigate.addListener(onBeforeNavigate);
+    chrome.webNavigation.onCommitted.addListener(onCommitted);
+    chrome.tabs.onUpdated.addListener(onUpdated);
 
     if ( typeof this.onClosed === 'function' ) {
         chrome.tabs.onRemoved.addListener(this.onClosed);
     }
 
-    if ( typeof this.onPopup === 'function' ) {
-        chrome.webNavigation.onCreatedNavigationTarget.addListener(onCreatedNavigationTarget);
-    }
 };
 
 /******************************************************************************/
@@ -115,7 +208,11 @@ vAPI.tabs.get = function(tabId, callback) {
         if ( typeof tabId === 'string' ) {
             tabId = parseInt(tabId, 10);
         }
-        chrome.tabs.get(tabId, onTabReady);
+        if ( typeof tabId !== 'number' || isNaN(tabId) ) {
+            onTabReady(null);
+        } else {
+            chrome.tabs.get(tabId, onTabReady);
+        }
         return;
     }
     var onTabReceived = function(tabs) {
@@ -222,6 +319,33 @@ vAPI.tabs.open = function(details) {
 
 /******************************************************************************/
 
+// Replace the URL of a tab. Noop if the tab does not exist.
+
+vAPI.tabs.replace = function(tabId, url) {
+    var targetURL = url;
+
+    // extension pages
+    if ( /^[\w-]{2,}:/.test(targetURL) !== true ) {
+        targetURL = vAPI.getURL(targetURL);
+    }
+
+    if ( typeof tabId !== 'number' ) {
+        tabId = parseInt(tabId, 10);
+        if ( isNaN(tabId) ) {
+            return;
+        }
+    }
+
+    chrome.tabs.update(tabId, { url: targetURL }, function() {
+        // this prevent console error
+        if ( chrome.runtime.lastError ) {
+            return;
+        }
+    });
+};
+
+/******************************************************************************/
+
 vAPI.tabs.remove = function(tabId) {
     var onTabRemoved = function() {
         if ( vAPI.lastError() ) {
@@ -262,8 +386,8 @@ vAPI.tabs.injectScript = function(tabId, details, callback) {
 
 // Must read: https://code.google.com/p/chromium/issues/detail?id=410868#c8
 
-// https://github.com/gorhill/uBlock/issues/19
-// https://github.com/gorhill/uBlock/issues/207
+// https://github.com/chrisaljoudi/uBlock/issues/19
+// https://github.com/chrisaljoudi/uBlock/issues/207
 // Since we may be called asynchronously, the tab id may not exist
 // anymore, so this ensures it does still exist.
 
@@ -428,7 +552,7 @@ CallbackWrapper.prototype.init = function(port, request) {
 };
 
 CallbackWrapper.prototype.proxy = function(response) {
-    // https://github.com/gorhill/uBlock/issues/383
+    // https://github.com/chrisaljoudi/uBlock/issues/383
     if ( this.messaging.ports.hasOwnProperty(this.port.name) ) {
         this.port.postMessage({
             requestId: this.request.requestId,
@@ -465,7 +589,7 @@ vAPI.net.registerListeners = function() {
         var tail = µburi.path.slice(-6);
         var pos = tail.lastIndexOf('.');
 
-        // https://github.com/gorhill/uBlock/issues/862
+        // https://github.com/chrisaljoudi/uBlock/issues/862
         // If no transposition possible, transpose to `object` as per
         // Chromium bug 410382 (see below)
         if ( pos === -1 ) {
@@ -521,56 +645,6 @@ vAPI.net.registerListeners = function() {
         },
         this.onHeadersReceived.extra
     );
-
-    // Intercept root frame requests.
-    // This is where we identify and block popups early, whenever possible.
-
-    var onBeforeSendHeaders = function(details) {
-        // Do not block behind the scene requests.
-        if ( vAPI.isNoTabId(details.tabId) ) {
-            return;
-        }
-
-        // Only root document.
-        if ( details.parentFrameId !== -1 ) {
-            return;
-        }
-
-        var referrer = headerValue(details.requestHeaders, 'referer');
-        if ( referrer === '' ) {
-            return;
-        }
-
-        var result = vAPI.tabs.onPopup({
-            openerTabId: undefined,
-            openerURL: referrer,
-            targetTabId: details.tabId,
-            targetURL: details.url
-        });
-
-        if ( result ) {
-            return { 'cancel': true };
-        }
-    };
-
-    var headerValue = function(headers, name) {
-        var i = headers.length;
-        while ( i-- ) {
-            if ( headers[i].name.toLowerCase() === name ) {
-                return headers[i].value;
-            }
-        }
-        return '';
-    };
-
-    chrome.webRequest.onBeforeSendHeaders.addListener(
-        onBeforeSendHeaders,
-        {
-            'urls': [ 'http://*/*', 'https://*/*' ],
-            'types': [ 'main_frame' ]
-        },
-        [ 'blocking', 'requestHeaders' ]
-    );
 };
 
 /******************************************************************************/
@@ -635,8 +709,9 @@ vAPI.onLoadAllCompleted = function() {
         var i = tabs.length, tab;
         while ( i-- ) {
             tab = tabs[i];
-            µb.bindTabToPageStats(tab.id, tab.url);
-            // https://github.com/gorhill/uBlock/issues/129
+            µb.tabContextManager.commit(tab.id, tab.url);
+            µb.bindTabToPageStats(tab.id);
+            // https://github.com/chrisaljoudi/uBlock/issues/129
             scriptStart(tab.id);
         }
     };
